"""
TA has moved into RRT. These tests remain here for historical reasons.
"""

import numpy.random as rnd
from numpy.testing import assert_

from alns.accept import RecordToRecordTravel
from alns.tests.states import One, Zero


def test_accepts_better():
    threshold_accept = RecordToRecordTravel(1, 0, 0.1, cmp_best=False)
    assert_(threshold_accept(rnd.RandomState(), Zero(), One(), Zero()))


def test_rejects_worse():
    threshold_accept = RecordToRecordTravel(0.5, 0.2, 0.1)

    # This results in a relative worsening of plus one, which is bigger than
    # the threshold (0.5).
    assert_(not threshold_accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_accepts_equal():
    threshold_accept = RecordToRecordTravel(0, 0, 0.1, cmp_best=False)

    # Even at the strictest threshold, this should be accepted since the
    # relative improvement is zero (they are equal).
    assert_(threshold_accept(rnd.RandomState(), Zero(), Zero(), Zero()))


def test_linear_threshold_update():
    threshold_accept = RecordToRecordTravel(5, 0, 1, cmp_best=False)

    # For the first five, the threshold is, resp., 5, 4, 3, 2, 1. The relative
    # worsening is plus one, so this should be accepted (lower or equal to
    # threshold).
    for _ in range(5):
        assert_(threshold_accept(rnd.RandomState(), Zero(), Zero(), One()))

    # Threshold is now zero, so this should no longer be accepted.
    assert_(not threshold_accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_exponential_threshold_update():
    threshold_accept = RecordToRecordTravel(5, 0, 0.1, "exponential", False)

    # The relative worsening is plus one, so this should be accepted initially,
    # as the threshold is 5, resp. 0.5. In the second case, 1 > 0.5, so the
    # second should be rejected.
    assert_(threshold_accept(rnd.RandomState(), Zero(), Zero(), One()))
    assert_(not threshold_accept(rnd.RandomState(), Zero(), Zero(), One()))



import logging
import time
from collections import defaultdict
from typing import Callable, Dict, Iterable, List, Optional, Tuple

import numpy as np
import numpy.random as rnd

from alns.Result import Result
from alns.State import State
from alns.Statistics import Statistics
from alns.accept import AcceptanceCriterion
from alns.stop import StoppingCriterion
from alns.weights import WeightScheme

# Potential candidate solution consideration outcomes.
_BEST = 0
_BETTER = 1
_ACCEPT = 2
_REJECT = 3

# TODO this should become a Protocol to allow for kwargs. See also this issue:
#  https://stackoverflow.com/q/61569324/4316405.
_OperatorType = Callable[[State, rnd.RandomState], State]

logger = logging.getLogger(__name__)


class ALNS:
    """
    Implements the adaptive large neighbourhood search (ALNS) algorithm.
    The implementation optimises for a minimisation problem, as explained
    in the text by Pisinger and Røpke (2010).

    Parameters
    ----------
    rnd_state
        Optional random state to use for random number generation. When
        passed, this state is used for operator selection and general
        computations requiring random numbers. It is also passed to the
        destroy and repair operators, as a second argument.

    References
    ----------
    [1]: Pisinger, D., and Røpke, S. (2010). Large Neighborhood Search. In
         M. Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399
         - 420). Springer.
    """

    def __init__(self, rnd_state: rnd.RandomState = rnd.RandomState()):
        self._d_ops: Dict[str, _OperatorType] = {}
        self._r_ops: Dict[str, _OperatorType] = {}
        self._only_after: Dict[_OperatorType, set] = defaultdict(set)

        # Optional callback that may be used to improve a new best solution
        # further, via e.g. local search.
        self._on_best: Optional[_OperatorType] = None

        self._rnd_state = rnd_state

    @property
    def destroy_operators(self) -> List[Tuple[str, _OperatorType]]:
        """
        Returns the destroy operators set for the ALNS algorithm.

        Returns
        -------
        A list of (name, operator) tuples. Their order is the same as the one
        in which they were passed to the ALNS instance.
        """
        return list(self._d_ops.items())

    @property
    def repair_operators(self) -> List[Tuple[str, _OperatorType]]:
        """
        Returns the repair operators set for the ALNS algorithm.

        Returns
        -------
        A list of (name, operator) tuples. Their order is the same as the one
        in which they were passed to the ALNS instance.
        """
        return list(self._r_ops.items())

    def add_destroy_operator(self, op: _OperatorType, name: str = None):
        """
        Adds a destroy operator to the heuristic instance.

        Parameters
        ----------
        op
            An operator that, when applied to the current state, returns a new
            state reflecting its implemented destroy action. The second
            argument is the random state constructed from the passed-in seed.
        name
            Optional name argument, naming the operator. When not passed, the
            function name is used instead.
        """
        logger.debug(f"Adding destroy operator {op.__name__}.")
        self._d_ops[op.__name__ if name is None else name] = op

    def add_repair_operator(
        self,
        op: _OperatorType,
        name: str = None,
        *,
        only_after: Optional[Iterable[_OperatorType]] = None,
    ):
        """
        Adds a repair operator to the heuristic instance.

        Parameters
        ----------
        op
            An operator that, when applied to the destroyed state, returns a
            new state reflecting its implemented repair action. The second
            argument is the random state constructed from the passed-in seed.
        name
            Optional name argument, naming the operator. When not passed, the
            function name is used instead.
        only_after
            Optional keyword-only argument indicating which destroy operators
            work with the passed-in repair operator. If passed, this argument
            should be an iterable (e.g. a list) of destroy operators. If not
            passed, the default is to assume that all destroy operators work
            with the new repair operator.
        """
        logger.debug(f"Adding repair operator {op.__name__}.")
        self._r_ops[name if name else op.__name__] = op

        if only_after:
            self._only_after[op].update(only_after)

    def _compute_op_coupling(self) -> np.ndarray:
        """
        Internal helper to compute a matrix that describes the
        coupling between destroy and repair operators. The matrix has size
        |d_ops|-by-|r_ops| and entry (i, j) is 1 if destroy operator i can
        be used in conjunction with repair operator j and 0 otherwise.

        If the only_after keyword-only argument was not used when adding
        the repair operators, then all entries of the matrix are 1.
        """
        op_coupling = np.ones((len(self._d_ops), len(self._r_ops)))

        for r_idx, (_, r_op) in enumerate(self.repair_operators):
            coupled_d_ops = self._only_after[r_op]

            for d_idx, (_, d_op) in enumerate(self.destroy_operators):
                if coupled_d_ops and d_op not in coupled_d_ops:
                    op_coupling[d_idx, r_idx] = 0

        # Destroy operators must be coupled with at least one repair operator
        d_idcs = np.flatnonzero(np.count_nonzero(op_coupling, axis=1) == 0)

        if d_idcs.size != 0:
            d_name, _ = self.destroy_operators[d_idcs[0]]
            raise ValueError(f"{d_name} has no coupled repair operators.")

        return op_coupling

    def iterate(
        self,
        initial_solution: State,
        weight_scheme: WeightScheme,
        crit: AcceptanceCriterion,
        stop: StoppingCriterion,
        **kwargs,
    ) -> Result:
        """
        Runs the adaptive large neighbourhood search heuristic [1], using the
        previously set destroy and repair operators. The first solution is set
        to the passed-in initial solution, and then subsequent solutions are
        computed by iteratively applying the operators.

        Parameters
        ----------
        initial_solution
            The initial solution, as a State object.
        weight_scheme
            The weight scheme to use for updating the (adaptive) weights. See
            also the ``alns.weight_schemes`` module for an overview.
        crit
            The acceptance criterion to use for candidate states. See also
            the ``alns.criteria`` module for an overview.
        stop
            The stopping criterion to use for stopping the iterations.
            See also the ``alns.stopping_criteria`` module for an overview.

        **kwargs
            Optional keyword arguments. These are passed to the operators,
            including callbacks.

        Raises
        ------
        ValueError
            When the parameters do not meet requirements.

        Returns
        -------
        A result object, containing the best solution and some additional
        statistics.

        References
        ----------
        [1]: Pisinger, D., & Røpke, S. (2010). Large Neighborhood Search. In M.
        Gendreau (Ed.), *Handbook of Metaheuristics* (2 ed., pp. 399-420).
        Springer.

        [2]: S. Røpke and D. Pisinger (2006). A unified heuristic for a large
        class of vehicle routing problems with backhauls. *European Journal of
        Operational Research*, 171: 750–775, 2006.
        """
        if len(self.destroy_operators) == 0 or len(self.repair_operators) == 0:
            raise ValueError("Missing destroy or repair operators.")

        curr = best = initial_solution
        init_obj = initial_solution.objective()
        op_coupling = self._compute_op_coupling()

        logger.debug(f"Initial solution has objective {init_obj:.2f}.")

        stats = Statistics()
        stats.collect_objective(init_obj)
        stats.collect_runtime(time.perf_counter())

        while not stop(self._rnd_state, best, curr):
            d_idx, r_idx = weight_scheme.select_operators(
                self._rnd_state, op_coupling
            )

            d_name, d_operator = self.destroy_operators[d_idx]
            r_name, r_operator = self.repair_operators[r_idx]

            logger.debug(f"Selected operators {d_name} and {r_name}.")

            destroyed = d_operator(curr, self._rnd_state, **kwargs)
            cand = r_operator(destroyed, self._rnd_state, **kwargs)

            best, curr, s_idx = self._eval_cand(
                crit, best, curr, cand, **kwargs
            )

            weight_scheme.update_weights(d_idx, r_idx, s_idx)

            stats.collect_objective(curr.objective())
            stats.collect_destroy_operator(d_name, s_idx)
            stats.collect_repair_operator(r_name, s_idx)
            stats.collect_runtime(time.perf_counter())

        logger.info(f"Finished iterating in {stats.total_runtime:.2f}s.")

        return Result(best, stats)

    def on_best(self, func: _OperatorType):
        """
        Sets a callback function to be called when ALNS finds a new global best
        solution state.

        Parameters
        ----------
        func
            A function that should take a solution State as its first argument,
            and a numpy RandomState as its second (cf. the operator signature).
            It should return a (new) solution State.
        """
        logger.debug(f"Adding on_best callback {func.__name__}.")
        self._on_best = func

    def _eval_cand(
        self,
        crit: AcceptanceCriterion,
        best: State,
        curr: State,
        cand: State,
        **kwargs,
    ) -> Tuple[State, State, int]:
        """
        Considers the candidate solution by comparing it against the best and
        current solutions. Candidate solutions are accepted based on the
        passed-in acceptance criterion. The (possibly new) best and current
        solutions are returned, along with a weight index (best, better,
        accepted, rejected).

        Returns
        -------
        A tuple of the best and current solution, along with the weight index.
        """
        w_idx = _REJECT

        if crit(self._rnd_state, best, curr, cand):  # accept candidate
            w_idx = _BETTER if cand.objective() < curr.objective() else _ACCEPT
            curr = cand

        if cand.objective() < best.objective():  # candidate is new best
            logger.info(f"New best with objective {cand.objective():.2f}.")

            if self._on_best:
                cand = self._on_best(cand, self._rnd_state, **kwargs)

            return cand, cand, _BEST

        return best, curr, w_idx



from collections import deque

from alns.accept.AcceptanceCriterion import AcceptanceCriterion


class LateAcceptanceHillClimbing(AcceptanceCriterion):
    """
    The Late Acceptance Hill Climbing (LAHC) criterion accepts a candidate
    solution when it is better than the current solution from a number of
    iterations ago.

    This implementation is based on the description of LAHC in [1].

    Parameters
    ----------
    lookback_period: int
        Non-negative integer specifying which solution to compare against
        for late acceptance. In particular, LAHC compares against the
        then-current solution from `lookback_period` iterations ago.
        If set to 0, then LAHC reverts to regular hill climbing.
    greedy: bool
        If set, LAHC always accepts a candidate that is better than the
        current solution.
    better_history: bool
        If set, LAHC uses a history management strategy where current solutions
        are stored only if they improve the then-current solution from
        `lookback_period` iterations ago. Otherwise, the then-current solution
        is stored again.

    References
    ----------
    [1]: Burke, E. K., & Bykov, Y. The late acceptance hill-climbing heuristic.
         *European Journal of Operational Research* (2017), 258(1), 70-78.
    """

    def __init__(
        self,
        lookback_period: int,
        greedy: bool = False,
        better_history: bool = False,
    ):
        self._lookback_period = lookback_period
        self._greedy = greedy
        self._better_history = better_history

        if lookback_period < 0:
            raise ValueError("lookback_period must be a non-negative integer.")

        self._history: deque = deque([], maxlen=lookback_period)

    @property
    def lookback_period(self):
        return self._lookback_period

    @property
    def greedy(self):
        return self._greedy

    @property
    def better_history(self):
        return self._better_history

    def __call__(self, rnd, best, current, candidate):
        if not self._history:
            self._history.append(current.objective())
            return candidate.objective() < current.objective()

        res = candidate.objective() < self._history[0]

        if not res and self._greedy:
            res = candidate.objective() < current.objective()

        if self._better_history:
            self._history.append(min(current.objective(), self._history[0]))
        else:
            self._history.append(current.objective())

        return res



from typing import List

import numpy as np
from numpy.testing import assert_almost_equal, assert_raises
from pytest import mark

from alns.weights import SegmentedWeights


@mark.parametrize("seg_decay", [1.01, -0.01, -0.5, 1.5])
def test_raises_invalid_seg_decay(seg_decay: float):
    with assert_raises(ValueError):
        SegmentedWeights([0, 0, 0, 0], 1, 1, seg_decay)


@mark.parametrize("seg_decay", np.linspace(0, 1, num=5))
def test_does_not_raise_valid_seg_decay(seg_decay: float):
    SegmentedWeights([0, 0, 0, 0], 1, 1, seg_decay)


@mark.parametrize(
    "scores,seg_decay,expected",
    [
        ([5, 3, 2, 1], 0, [3, 3]),  # 1 * 0 + (0 + 3) * 1 = 3
        ([5, 3, 2, 1], 1, [1, 1]),  # 1 * 1 + (0 + 3) * 0 = 1
        ([5, 3, 2, 1], 0.5, [2, 2]),  # .5 * 1 + (0 + 3) * .5 = 2
        ([5, 5, 5, 5], 0, [5, 5]),  # etc. etc.
        ([5, 5, 5, 5], 1, [1, 1]),
        ([5, 5, 5, 5], 0.5, [3, 3]),
    ],
)
def test_update_weights(
    scores: List[float], seg_decay: float, expected: List[float]
):
    rnd_state = np.random.RandomState(1)
    weights = SegmentedWeights(scores, 1, 1, seg_decay, 1)
    op_coupling = np.ones((1, 1))

    # TODO other weights?
    weights.update_weights(0, 0, 1)
    weights.select_operators(rnd_state, op_coupling)

    assert_almost_equal(weights.destroy_weights[0], expected[0])
    assert_almost_equal(weights.repair_weights[0], expected[1])


@mark.parametrize(
    "scores,num_destroy,num_repair,seg_decay,seg_length",
    [
        ([5, 3, 2, -1], 1, 1, 0, 1),  # negative score
        ([5, 3, 2], 1, 1, 0, 1),  # len(score) < 4
        ([5, 3, 2, 1], 0, 1, 0, 1),  # no destroy operator
        ([5, 3, 2, 1], 1, 0, 0, 1),  # no repair operator
        ([5, 3, 2, 1], 1, 1, -1, 1),  # seg_decay < 0
        ([5, 3, 2, 1], 1, 1, 2, 1),  # seg_decay > 1
        ([5, 3, 2, 1], 1, 1, 0.5, 0),  # seg_length < 1
    ],
)
def test_raises_invalid_arguments(
    scores: List[float],
    num_destroy: int,
    num_repair: int,
    seg_decay: float,
    seg_length: int,
):
    with assert_raises(ValueError):
        SegmentedWeights(
            scores, num_destroy, num_repair, seg_decay, seg_length
        )


# TODO test select weights, at iteration start



from .AcceptanceCriterion import AcceptanceCriterion
from .GreatDeluge import GreatDeluge
from .HillClimbing import HillClimbing
from .LateAcceptanceHillClimbing import LateAcceptanceHillClimbing
from .NonLinearGreatDeluge import NonLinearGreatDeluge
from .RandomWalk import RandomWalk
from .RecordToRecordTravel import RecordToRecordTravel
from .SimulatedAnnealing import SimulatedAnnealing
from .WorseAccept import WorseAccept



from numpy.random import RandomState

from alns.State import State
from alns.stop.StoppingCriterion import StoppingCriterion


class MaxIterations(StoppingCriterion):
    """
    Criterion that stops after a maximum number of iterations.
    """

    def __init__(self, max_iterations: int):
        if max_iterations < 0:
            raise ValueError("max_iterations < 0 not understood.")

        self._max_iterations = max_iterations
        self._current_iteration = 0

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def __call__(self, rnd: RandomState, best: State, current: State) -> bool:
        self._current_iteration += 1

        return self._current_iteration > self.max_iterations



import logging

from alns.accept.AcceptanceCriterion import AcceptanceCriterion
from alns.accept.update import update

logger = logging.getLogger(__name__)


class RecordToRecordTravel(AcceptanceCriterion):
    """
    The Record-to-Record Travel (RRT) criterion accepts a candidate solution
    if the absolute gap between the candidate and the best or current solution
    is smaller than a threshold. The threshold is updated in each iteration as:

    ``threshold = max(end_threshold, threshold - step)`` (linear)

    ``threshold = max(end_threshold, step * threshold)`` (exponential)

    where the initial threshold is set to ``start_threshold``.

    Parameters
    ----------
    start_threshold
        The initial threshold.
    end_threshold
        The final threshold.
    step
        The updating step.
    method
        The updating method, one of {'linear', 'exponential'}. Default
        'linear'.
    cmp_best
        This parameter determines whether we use default RRT (True), or
        threshold accepting (False). By default, `cmp_best` is True, in which
        case RRT checks whether the difference between the candidate and best
        solution is below the threshold [2]. If `cmp_best` is False, RRT takes
        the difference between the candidate and current solution instead. This
        yields the behaviour of threshold accepting (TA), see [3] for details.

    References
    ----------
    [1]: Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
         criteria for the adaptive large neighbourhood search metaheuristic.
         *Journal of Heuristics* (2018) 24 (5): 783–815.
    [2]: Dueck, G. New optimization heuristics: The great deluge algorithm and
         the record-to-record travel. *Journal of Computational Physics* (1993)
         104 (1): 86-92.
    [3]: Dueck, G., Scheuer, T. Threshold accepting: A general purpose
         optimization algorithm appearing superior to simulated annealing.
         *Journal of Computational Physics* (1990) 90 (1): 161-175.
    """

    def __init__(
        self,
        start_threshold: float,
        end_threshold: float,
        step: float,
        method: str = "linear",
        cmp_best: bool = True,
    ):
        if start_threshold < 0 or end_threshold < 0 or step < 0:
            raise ValueError("Thresholds and step must be non-negative.")

        if start_threshold < end_threshold:
            raise ValueError("start_threshold < end_threshold not understood.")

        if method == "exponential" and step > 1:
            raise ValueError("Exponential updating cannot have step > 1.")

        if method not in ["linear", "exponential"]:
            raise ValueError("Method must be one of ['linear', 'exponential']")

        self._start_threshold = start_threshold
        self._end_threshold = end_threshold
        self._step = step
        self._method = method
        self._cmp_best = cmp_best

        self._threshold = start_threshold

    @property
    def start_threshold(self) -> float:
        return self._start_threshold

    @property
    def end_threshold(self) -> float:
        return self._end_threshold

    @property
    def step(self) -> float:
        return self._step

    @property
    def method(self) -> str:
        return self._method

    def __call__(self, rnd, best, current, candidate):
        # From [2] p. 87 (RRT; best), and [3] p. 162 (TA; current).
        baseline = best if self._cmp_best else current
        res = candidate.objective() - baseline.objective() <= self._threshold

        self._threshold = max(
            self.end_threshold, update(self._threshold, self.step, self.method)
        )

        return res

    @classmethod
    def autofit(
        cls,
        init_obj: float,
        start_gap: float,
        end_gap: float,
        num_iters: int,
        method: str = "linear",
    ):
        """
        Returns an RRT object such that the start threshold is set at
        ``start_gap`` percent of the initial objective ``init_obj``
        and the end threshold is set at ``end_gap`` percent of the
        initial objective. The step parameter is then chosen such that
        the end threshold is reached in ``num_iters`` iterations using
        the passed-in update method.

        Parameters
        ----------
        init_obj
            The initial solution objective
        start_gap
            Percentage gap of the initial objective used for deriving
            the start threshold.
        end_gap
            Percentage gap of the initial objective used for deriving
            the end threshold.
        num_iters
            The number of iterations that the ALNS algorithm will run.
        method
            Updating method. Default 'linear'.

        Returns
        -------
        An autofitted RecordToRecordTravel acceptance criterion.
        """
        if not (0 <= end_gap <= start_gap):
            raise ValueError("Must have 0 <= end_gap <= start_gap")

        if num_iters <= 0:
            raise ValueError("Non-positive num_iters not understood.")

        if method not in ["linear", "exponential"]:
            raise ValueError("Method must be one of ['linear', 'exponential']")

        start_threshold = start_gap * init_obj
        end_threshold = end_gap * init_obj

        if method == "linear":
            step = (start_threshold - end_threshold) / num_iters
        elif method == "exponential":
            step = (end_threshold / start_threshold) ** (1 / num_iters)

        logger.info(
            f"Autofit start_threshold {start_threshold:.2f}"
            f"end_threshold {end_threshold:.2f}, step {step:.2f}."
        )

        return cls(start_threshold, end_threshold, step, method=method)



from alns.accept.AcceptanceCriterion import AcceptanceCriterion


class RandomWalk(AcceptanceCriterion):
    """
    The random walk criterion always accepts the candidate solution.
    """

    def __call__(self, rnd, best, current, candidate):
        return True



from collections import defaultdict
from typing import DefaultDict, List

import numpy as np


class Statistics:
    """
    Statistics object that stores some iteration results. Populated by the ALNS
    algorithm.
    """

    def __init__(self):
        self._objectives = []
        self._runtimes = []

        self._destroy_operator_counts = defaultdict(lambda: [0, 0, 0, 0])
        self._repair_operator_counts = defaultdict(lambda: [0, 0, 0, 0])

    @property
    def objectives(self) -> np.ndarray:
        """
        Returns an array of previous objective values, tracking progress.
        """
        return np.array(self._objectives)

    @property
    def start_time(self) -> float:
        """
        Return the reference start time to compute the runtimes.
        """
        return self._runtimes[0]

    @property
    def total_runtime(self) -> float:
        """
        Return the total runtime (in seconds).
        """
        return self._runtimes[-1] - self._runtimes[0]

    @property
    def runtimes(self) -> np.ndarray:
        """
        Returns an array of iteration run times (in seconds).
        """
        return np.diff(self._runtimes)

    @property
    def destroy_operator_counts(self) -> DefaultDict[str, List[float]]:
        """
        Returns the destroy operator counts, as a dictionary of operator names
        to lists of counts. Such a list consists of four elements, one for
        each possible outcome, and counts the number of times that the
        application of that operator resulted in such an outcome.

        Returns
        -------
        Destroy operator counts.
        """
        return self._destroy_operator_counts

    @property
    def repair_operator_counts(self) -> DefaultDict[str, List[float]]:
        """
        Returns the repair operator counts, as a dictionary of operator names
        to lists of counts. Such a list consists of four elements, one for
        each possible outcome, and counts the number of times that the
        application of that operator resulted in such an outcome.

        Returns
        -------
        Repair operator counts.
        """
        return self._repair_operator_counts

    def collect_objective(self, objective: float):
        """
        Collects an objective value.

        Parameters
        ----------
        objective
            The objective value to be collected.
        """
        self._objectives.append(objective)

    def collect_runtime(self, time: float):
        """
        Collects the time one iteration took.

        Parameters
        ----------
        time
            Time in seconds.
        """
        self._runtimes.append(time)

    def collect_destroy_operator(self, operator_name: str, s_idx: int):
        """
        Collects a score (index) for a used destroy operator. This maintains
        count of the number of times this operator was used, and what result
        came from its use.

        Parameters
        ----------
        operator_name
            Operator name. This was set when the operator was passed to the
            ALNS instance.
        s_idx
            Score indices used for the various iteration outcomes.
        """
        self._destroy_operator_counts[operator_name][s_idx] += 1

    def collect_repair_operator(self, operator_name: str, s_idx: int):
        """
        Collects a score (index) for a used repair operator. This maintains
        count of the number of times this operator was used, and what result
        came from its use.

        Parameters
        ----------
        operator_name
            Operator name. This was set when the operator was passed to the
            ALNS instance.
        s_idx
            Score indices used for the various iteration outcomes.
        """
        self._repair_operator_counts[operator_name][s_idx] += 1



import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.accept import NonLinearGreatDeluge
from alns.tests.states import One, Two, VarObj, Zero


@mark.parametrize(
    "alpha, beta, gamma, delta",
    [
        (1, 0.5, 1, 1),  # alpha cannot be <= 1
        (2, 1, 1, 1),  # beta cannot be outside of (0, 1)
        (2, 0.5, 0, 1),  # gamma cannot be <= 0
        (2, 0.5, 1, 0),  # delta cannot be <= 0
    ],
)
def test_raise_invalid_parameters(alpha, beta, gamma, delta):
    with assert_raises(ValueError):
        NonLinearGreatDeluge(alpha, beta, gamma, delta)


@mark.parametrize(
    "alpha,beta,gamma,delta",
    [(2, 0.5, 1, 1), (2, 0.5, 0.5, 0.5), (1.1, 0.01, 0.25, 0.25)],
)
def test_no_raise_valid_parameters(alpha, beta, gamma, delta):
    NonLinearGreatDeluge(alpha, beta, gamma, delta)


@mark.parametrize(
    "alpha,beta,gamma,delta",
    [(2, 0.5, 1, 1), (2, 0.5, 0.5, 0.5), (1.1, 0.01, 0.25, 0.25)],
)
def test_properties(alpha, beta, gamma, delta):
    nlgd = NonLinearGreatDeluge(alpha, beta, gamma, delta)

    assert_equal(nlgd.alpha, alpha)
    assert_equal(nlgd.beta, beta)
    assert_equal(nlgd.gamma, gamma)
    assert_equal(nlgd.delta, delta)


def test_raise_zero_best():
    """
    Test if an error is raised when the initial solution has value 0.
    The initial solution is used to compute the initial threshold. An initial
    threshold with value 0 cannot be updated because the relative gap in the
    update method cannot be computed.
    """
    nlgd = NonLinearGreatDeluge(2, 0.1, 0.01, 1)

    with assert_raises(ValueError):
        nlgd(rnd.RandomState(), Zero(), One(), One())


def test_accepts_below_threshold():
    nlgd = NonLinearGreatDeluge(1.01, 0.5, 0.01, 1)

    # Initial threshold is set at 1.01, hence One should be accepted.
    assert_(nlgd(rnd.RandomState(), One(), Zero(), One()))


def test_rejects_above_threshold():
    nlgd = NonLinearGreatDeluge(1.99, 0.5, 1, 1)

    # Initial threshold is set at 1.99 and the current solution is Zero,
    # hence Two should be rejected.
    assert_(not nlgd(rnd.RandomState(), One(), Zero(), Two()))


def test_rejects_equal_threshold():
    nlgd = NonLinearGreatDeluge(2, 0.5, 1, 1)

    # Initial threshold is set at 2 and the current solution is Zero,
    # hence Two should be rejected.
    assert_(not nlgd(rnd.RandomState(), One(), Zero(), Two()))


def test_accepts_improving_current():
    nlgd = NonLinearGreatDeluge(2, 0.5, 1, 1)

    # Candidate is not below the threshold (2 == 2) but does improve the
    # current solution (2 < 3), hence candidate should be accepted.
    assert_(nlgd(rnd.RandomState(), One(), VarObj(3), Two()))


def test_evaluate_consecutive_solutions():
    """
    Test if NLGD correctly accepts and rejects consecutive solutions.
    """
    nlgd = NonLinearGreatDeluge(1.5, 0.1, 0.02, 2)

    # The first is above the threshold (2 > 1.5) so the first should be
    # rejected. The threshold is then linearly increased to 1.51. The second
    # candidate is below the threshold (0 < 1.51), so the second should be
    # accepted. The threshold is then exponentially decreased to 1.20. The
    # third candidate is below the theshold (1 < 1.20) and is accepted.
    assert_(not nlgd(rnd.RandomState(), One(), Zero(), Two()))
    assert_(nlgd(rnd.RandomState(), One(), Zero(), Zero()))
    assert_(nlgd(rnd.RandomState(), One(), Zero(), One()))



import numpy.random as rnd
from numpy.testing import assert_

from alns.accept import RandomWalk
from alns.tests.states import One, Zero


def test_accepts_better():
    """
    Tests if the random walk method accepts a better solution.
    """
    random_walk = RandomWalk()
    assert_(random_walk(rnd.RandomState(), One(), One(), Zero()))


def test_accepts_worse():
    """
    Tests if the random walk method accepts a worse solution.
    """
    random_walk = RandomWalk()
    assert_(random_walk(rnd.RandomState(), Zero(), Zero(), One()))


def test_accepts_equal():
    """
    Tests if the random walk method accepts a solution that results in the
    same objective value.
    """
    random_walk = RandomWalk()
    assert_(random_walk(rnd.RandomState(), Zero(), Zero(), Zero()))



import pytest
from numpy.random import RandomState
from numpy.testing import assert_, assert_equal, assert_raises

from alns.stop import NoImprovement
from alns.tests.states import One, Zero


@pytest.mark.parametrize("max_iterations", [-10, -100, -1000])
def test_raise_negative_parameters(max_iterations: int):
    """
    max_iterations cannot be negative.
    """
    with assert_raises(ValueError):
        NoImprovement(max_iterations)


@pytest.mark.parametrize("max_iterations", [0, 10, 100, 1000])
def test_max_iterations(max_iterations):
    """
    Test if the max_iterations parameter is correctly set.
    """
    stop = NoImprovement(max_iterations)
    assert_equal(stop.max_iterations, max_iterations)


def test_zero_max_iterations():
    """
    Test if setting max_iterations to zero always stops.
    """
    stop = NoImprovement(0)
    rnd = RandomState()

    assert_(stop(rnd, One(), Zero()))
    assert_(stop(rnd, Zero(), Zero()))


def test_one_max_iterations():
    """
    Test if setting max_iterations to one only stops when a non-improving
    best solution has been found.
    """
    stop = NoImprovement(1)
    rnd = RandomState()

    assert_(not stop(rnd, One(), Zero()))
    assert_(not stop(rnd, Zero(), Zero()))
    assert_(stop(rnd, Zero(), Zero()))


@pytest.mark.parametrize("n", [10, 100, 1000])
def test_n_max_iterations_non_improving(n):
    """
    Test if setting max_iterations to n correctly stops with non-improving
    solutions. The first n iterations should not stop. Beyond that, the
    the criterion should stop.
    """
    stop = NoImprovement(n)
    rnd = RandomState()

    for _ in range(n):
        assert_(not stop(rnd, Zero(), Zero()))

    for _ in range(n):
        assert_(stop(rnd, Zero(), Zero()))


@pytest.mark.parametrize("n, k", [(10, 2), (100, 20), (1000, 200)])
def test_n_max_iterations_with_single_improvement(n, k):
    """
    Test if setting max_iterations to n correctly stops with a sequence
    of solutions, where the k-th solution is improving and the other solutions
    are non-improving. The first n+k-1 iterations should not stop. Beyond that,
    the criterion should stop.
    """
    stop = NoImprovement(n)
    rnd = RandomState()

    for _ in range(k):
        assert_(not stop(rnd, One(), Zero()))

    for _ in range(n):
        assert_(not stop(rnd, Zero(), Zero()))

    for _ in range(n):
        assert_(stop(rnd, Zero(), Zero()))



import numpy as np
import numpy.random as rnd
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_equal,
    assert_raises,
)
from pytest import mark

from alns import ALNS, State
from alns.accept import HillClimbing, SimulatedAnnealing
from alns.stop import MaxIterations, MaxRuntime
from alns.weights import SimpleWeights
from .states import One, Zero


# HELPERS ---------------------------------------------------------------------


def get_alns_instance(
    repair_operators=None, destroy_operators=None, seed=None
):
    """
    Test helper method.
    """
    alns = ALNS(rnd.RandomState(seed))

    if repair_operators is not None:
        for idx, repair_operator in enumerate(repair_operators):
            alns.add_repair_operator(repair_operator, name=str(idx))

    if destroy_operators is not None:
        for idx, destroy_operator in enumerate(destroy_operators):
            alns.add_destroy_operator(destroy_operator, name=str(idx))

    return alns


class ValueState(State):
    """
    Helper state for testing random values.
    """

    def __init__(self, value):
        self._value = value

    def objective(self):
        return self._value


def get_repair_operators(n):
    """
    Get a list of n dummy repair operators.
    """
    repair_operators = []

    for idx in range(n):
        op = lambda: None  # noqa: E731
        op.__name__ = f"Repair operator {idx}"
        repair_operators.append(op)

    return repair_operators


def get_destroy_operators(n):
    """
    Get a list of n dummy destroy operators.
    """
    destroy_operators = []

    for idx in range(n):
        op = lambda: None  # noqa: E731
        op.__name__ = f"Destroy operator {idx}"
        destroy_operators.append(op)

    return destroy_operators


# CALLBACKS -------------------------------------------------------------------


def test_on_best_is_called():
    """
    Tests if the callback is invoked when a new global best is found.
    """
    alns = get_alns_instance(
        [lambda state, rnd: Zero()], [lambda state, rnd: Zero()]
    )

    # Called when a new global best is found. In this case, that happens once:
    # in the only iteration below. It returns a state with value 10, which
    # should then also be returned by the entire algorithm.
    alns.on_best(lambda *args: ValueState(10))

    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)
    result = alns.iterate(One(), weights, HillClimbing(), MaxIterations(1))
    assert_equal(result.best_state.objective(), 10)


# OPERATORS -------------------------------------------------------------------


def test_add_destroy_operator():
    """
    Tests if adding a destroy operator correctly updates the number of
    operators available on the ALNS instance.
    """
    alns = ALNS()

    for count in [1, 2]:
        alns.add_destroy_operator(lambda state, rnd: state, name=str(count))
        assert_equal(len(alns.destroy_operators), count)


def test_add_destroy_operator_name():
    """
    Tests if adding a destroy operator without an explicit name correctly
    takes the function name instead.
    """

    def destroy_operator():  # placeholder
        pass

    alns = ALNS()

    alns.add_destroy_operator(destroy_operator)

    name, operator = alns.destroy_operators[0]

    assert_equal(name, "destroy_operator")
    assert_(operator is destroy_operator)


def test_add_repair_operator():
    """
    Tests if adding a repair operator correctly updates the number of
    operators available on the ALNS instance.
    """
    alns = ALNS()

    for count in [1, 2]:
        alns.add_repair_operator(lambda state, rnd: state, name=str(count))
        assert_equal(len(alns.repair_operators), count)


def test_add_repair_operator_name():
    """
    Tests if adding a repair operator without an explicit name correctly
    takes the function name instead.
    """

    def repair_operator():  # placeholder
        pass

    alns = ALNS()

    alns.add_repair_operator(repair_operator)

    name, operator = alns.repair_operators[0]

    assert_equal(name, "repair_operator")
    assert_(operator is repair_operator)


def test_compute_op_coupling():
    """
    Tests if the compute_op_coupling method correctly computes the matrix
    that describes the dependencies between repair and destroy operators.
    """
    alns = ALNS()

    d_operators = get_destroy_operators(2)
    r_operators = get_repair_operators(2)

    for d_op in d_operators:
        alns.add_destroy_operator(d_op)

    for r_op in r_operators:
        alns.add_repair_operator(r_op)

    op_coupling = alns._compute_op_coupling()

    assert_almost_equal(op_coupling, np.ones((2, 2)))


def test_compute_op_coupling_only_after():
    """
    Tests if the compute_op_coupling method correctly computes the matrix
    when the only_after paramter is specified for certain repair operators.
    """
    alns = ALNS()

    d_operators = get_destroy_operators(2)
    r_operators = get_repair_operators(2)

    for d_op in d_operators:
        alns.add_destroy_operator(d_op)

    for idx, r_op in enumerate(r_operators):
        alns.add_repair_operator(r_op, only_after=[d_operators[idx]])

    op_coupling = alns._compute_op_coupling()

    assert_almost_equal(op_coupling, np.eye(2))


def test_raise_uncoupled_destroy_op():
    """
    Tests if having a destroy operator that is not coupled to any of the
    repair operators raises an an error when computing the operator coupling.
    """
    alns = ALNS()

    d_operators = get_destroy_operators(2)
    r_operators = get_repair_operators(2)

    for d_op in d_operators:
        alns.add_destroy_operator(d_op)

    for r_op in r_operators:
        alns.add_repair_operator(r_op, only_after=[d_operators[0]])

    with assert_raises(ValueError):
        alns._compute_op_coupling()


# PARAMETERS ------------------------------------------------------------------


def test_raises_missing_destroy_operator():
    """
    Tests if the algorithm raises when no destroy operators have been set.
    """
    alns = get_alns_instance(repair_operators=[lambda state, rnd: None])

    # Pretend we have a destroy operator for the weight scheme, so that
    # does not raise an error.
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.95)

    with assert_raises(ValueError):
        alns.iterate(One(), weights, HillClimbing(), MaxIterations(1))


def test_raises_missing_repair_operator():
    """
    Tests if the algorithm raises when no repair operators have been set.
    """
    alns = get_alns_instance(destroy_operators=[lambda state, rnd: None])

    # Pretend we have a destroy operator for the weight scheme, so that
    # does not raise an error.
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.95)

    with assert_raises(ValueError):
        alns.iterate(One(), weights, HillClimbing(), MaxIterations(1))


def test_zero_max_iterations():
    """
    Test that the algorithm return the initial solution when the
    stopping criterion is zero max iterations.
    """
    alns = get_alns_instance(
        [lambda state, rnd: None], [lambda state, rnd: None]
    )

    initial_solution = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)

    result = alns.iterate(
        initial_solution, weights, HillClimbing(), MaxIterations(0)
    )

    assert_(result.best_state is initial_solution)


def test_zero_max_runtime():
    """
    Test that the algorithm return the initial solution when the
    stopping criterion is zero max runtime.
    """
    alns = get_alns_instance(
        [lambda state, rnd: None], [lambda state, rnd: None]
    )

    initial_solution = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)

    result = alns.iterate(
        initial_solution, weights, HillClimbing(), MaxRuntime(0)
    )

    assert_(result.best_state is initial_solution)


def test_iterate_kwargs_are_correctly_passed_to_operators():
    def test_operator(state, rnd, item):
        assert_(item is orig_item)
        return state

    alns = get_alns_instance([lambda state, rnd, item: state], [test_operator])

    init_sol = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)
    orig_item = object()

    alns.iterate(
        init_sol, weights, HillClimbing(), MaxIterations(10), item=orig_item
    )


def test_bugfix_pass_kwargs_to_on_best():
    """
    Exercises a bug where the on_best callback did not receive the kwargs
    passed to iterate().
    """

    def test_operator(state, rnd, item):
        assert_(item is orig_item)
        return Zero()  # better, so on_best is triggered

    alns = get_alns_instance([lambda state, rnd, item: state], [test_operator])
    alns.on_best(lambda state, rnd, item: state)

    init_sol = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)
    orig_item = object()

    alns.iterate(
        init_sol, weights, HillClimbing(), MaxIterations(10), item=orig_item
    )


# EXAMPLES --------------------------------------------------------------------


def test_trivial_example():
    """
    This tests the ALNS algorithm on a trivial example, where the initial
    solution is one, and any other operator returns zero.
    """
    alns = get_alns_instance(
        [lambda state, rnd: Zero()], [lambda state, rnd: Zero()]
    )

    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)
    result = alns.iterate(One(), weights, HillClimbing(), MaxIterations(100))

    assert_equal(result.best_state.objective(), 0)


@mark.parametrize("seed,desired", [(0, 0.01171), (1, 0.00011), (2, 0.01025)])
def test_fixed_seed_outcomes(seed: int, desired: float):
    """
    Tests if fixing a seed results in deterministic outcomes even when using a
    'random' acceptance criterion (here SA).
    """
    alns = get_alns_instance(
        [lambda state, rnd: ValueState(rnd.random_sample())],
        [lambda state, rnd: None],
        seed,
    )

    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)
    sa = SimulatedAnnealing(1, 0.25, 1 / 100)

    result = alns.iterate(One(), weights, sa, MaxIterations(100))
    assert_almost_equal(result.best_state.objective(), desired, decimal=5)


@mark.parametrize("max_iterations", [1, 10, 100])
def test_nonnegative_max_iterations(max_iterations):
    """
    Test that the result statistics have size equal to max iterations (+1).
    """
    alns = get_alns_instance(
        [lambda state, rnd: Zero()], [lambda state, rnd: Zero()]
    )

    initial_solution = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)

    result = alns.iterate(
        initial_solution,
        weights,
        HillClimbing(),
        MaxIterations(max_iterations),
    )

    assert_equal(len(result.statistics.objectives), max_iterations + 1)
    assert_equal(len(result.statistics.runtimes), max_iterations)


@mark.parametrize("max_runtime", [0.01, 0.05, 0.1])
def test_nonnegative_max_runtime(max_runtime):
    """
    Test that the result runtime statistics match the stopping criterion.
    """
    alns = get_alns_instance(
        [lambda state, rnd: Zero()], [lambda state, rnd: Zero()]
    )

    initial_solution = One()
    weights = SimpleWeights([1, 1, 1, 1], 1, 1, 0.5)

    result = alns.iterate(
        initial_solution, weights, HillClimbing(), MaxRuntime(max_runtime)
    )

    assert_almost_equal(
        sum(result.statistics.runtimes), max_runtime, decimal=3
    )


# TODO test more complicated examples?



from numpy.random import RandomState
from typing import Optional

from alns.State import State
from alns.stop.StoppingCriterion import StoppingCriterion


class NoImprovement(StoppingCriterion):
    """
    Criterion that stops if the best solution has not been improved
    after a number of iterations.

    Parameters
    ----------
    max_iterations
        The maximum number of non-improving iterations.
    """

    def __init__(self, max_iterations: int):
        if max_iterations < 0:
            raise ValueError("max_iterations < 0 not understood.")

        self._max_iterations = max_iterations
        self._target: Optional[float] = None
        self._counter = 0

    @property
    def max_iterations(self) -> int:
        return self._max_iterations

    def __call__(self, rnd: RandomState, best: State, current: State) -> bool:
        if self._target is None or best.objective() < self._target:
            self._target = best.objective()
            self._counter = 0
        else:
            self._counter += 1

        return self._counter >= self.max_iterations



from alns.accept.AcceptanceCriterion import AcceptanceCriterion
from alns.accept.update import update


class WorseAccept(AcceptanceCriterion):
    """
    The Worse Accept criterion accepts a candidate solution 1) if it improves
    over the current one or 2) with a given probability regardless of the cost.
    The probability is updated in each iteration as:

    ``prob = max(end_prob, prob - step)`` (linear)

    ``prob = max(end_prob, step * prob)`` (exponential)

    where the initial probability is set to ``start_prob``.

    Parameters
    ----------
    start_prob
        The initial probability.
    end_prob
        The final probability.
    step:
        The updating step.
    method
        The updating method, one of {'linear', 'exponential'}. Default
        'linear'.

    """

    def __init__(
        self,
        start_prob: float,
        end_prob: float,
        step: float,
        method: str = "linear",
    ):
        if not 0 <= end_prob <= start_prob <= 1:
            raise ValueError("Must have 0 <= start_prob <= end_prob <= 1")

        if step < 0:
            raise ValueError("Step cannot be negative.")

        if method == "exponential" and step > 1:
            raise ValueError("Exponential updating cannot have step > 1.")

        self._start_prob = start_prob
        self._end_prob = end_prob
        self._step = step
        self._method = method

        self._prob = start_prob

    @property
    def start_prob(self) -> float:
        return self._start_prob

    @property
    def end_prob(self) -> float:
        return self._end_prob

    @property
    def step(self) -> float:
        return self._step

    @property
    def method(self) -> str:
        return self._method

    def __call__(self, rnd, best, current, candidate):
        # Always accept better
        res = candidate.objective() < current.objective()

        if not res:  # maybe accept worse
            res = rnd.random() < self._prob

        self._prob = max(
            self.end_prob, update(self._prob, self.step, self.method)
        )

        return res



import numpy.random as rnd
from numpy.testing import assert_

from alns.accept import HillClimbing
from alns.tests.states import One, Zero


def test_accepts_better():
    """
    Tests if the hill climbing method accepts a better solution.
    """
    hill_climbing = HillClimbing()
    assert_(hill_climbing(rnd.RandomState(), One(), One(), Zero()))


def test_rejects_worse():
    """
    Tests if the hill climbing method accepts a worse solution.
    """
    hill_climbing = HillClimbing()
    assert_(not hill_climbing(rnd.RandomState(), Zero(), Zero(), One()))


def test_accepts_equal():
    """
    Tests if the hill climbing method accepts a solution that results in the
    same objective value.
    """
    hill_climbing = HillClimbing()
    assert_(hill_climbing(rnd.RandomState(), Zero(), Zero(), Zero()))



from numpy.testing import assert_equal, assert_almost_equal, assert_allclose

from alns.Statistics import Statistics


def test_empty_new_statistics():
    """
    Tests if a new Statistics object starts empty.
    """
    statistics = Statistics()

    assert_equal(len(statistics.objectives), 0)
    assert_equal(len(statistics.destroy_operator_counts), 0)
    assert_equal(len(statistics.repair_operator_counts), 0)


def test_collect_objectives():
    """
    Tests if a Statistics object properly collects objective values.
    """
    statistics = Statistics()

    for objective in range(1, 100):
        statistics.collect_objective(objective)

        assert_equal(len(statistics.objectives), objective)
        assert_almost_equal(statistics.objectives[-1], objective)


def test_collect_runtimes():
    """
    Tests if a Statistics object properly collects runtime values.
    """
    statistics = Statistics()

    for time in range(100):
        statistics.collect_runtime(time)

    # 100 - 1, since the first runtime is assumed to be the start time.
    assert_equal(len(statistics.runtimes), 99)
    assert_allclose(statistics.runtimes, 1)  # steps of one


def test_start_time():
    """
    Tests if the reference start time parameter is correctly set.
    """
    statistics = Statistics()

    for time in range(1):
        statistics.collect_runtime(time)

    assert_equal(statistics.start_time, 0)


def test_total_runtime():
    """
    Tests if the total runtime parameter is correctly set.
    """
    statistics = Statistics()

    for time in range(100):
        statistics.collect_runtime(time)

    assert_equal(statistics.total_runtime, 99)


def test_collect_destroy_counts_example():
    """
    Tests if collecting for a destroy operator works as expected in a simple
    example.
    """
    statistics = Statistics()

    # This should twice collect for a method "destroy_test", at index 1, such
    # that the total count is [0, 2, 0, 0] for this operator.
    statistics.collect_destroy_operator("destroy_test", 1)
    statistics.collect_destroy_operator("destroy_test", 1)

    for idx, count in enumerate([0, 2, 0, 0]):
        assert_equal(
            statistics.destroy_operator_counts["destroy_test"][idx], count
        )


def test_collect_repair_counts_example():
    """
    Tests if collecting for a repair operator works as expected in a simple
    example.
    """
    statistics = Statistics()

    # This should once collect for a method "repair_test", at index 2, such
    # that the total count is [0, 0, 1, 0] for this operator.
    statistics.collect_repair_operator("repair_test", 2)

    for idx, count in enumerate([0, 0, 1, 0]):
        assert_equal(
            statistics.repair_operator_counts["repair_test"][idx], count
        )



from typing import List

from .WeightScheme import WeightScheme


class SimpleWeights(WeightScheme):
    """
    A simple weighting scheme, where the operator weights are adjusted
    continuously throughout the algorithm runs. This works as follows.
    In each iteration, the old weight is updated with a score based on a
    convex combination of the existing weight and the new score, as:

    ``new_weight = op_decay * old_weight + (1 - op_decay) * score``

    Parameters
    ----------
    (other arguments are explained in ``WeightScheme``)

    op_decay
        Decay parameter in [0, 1]. This parameter is used to weigh the
        running performance of each operator.
    """

    def __init__(
        self,
        scores: List[float],
        num_destroy: int,
        num_repair: int,
        op_decay: float,
    ):
        super().__init__(scores, num_destroy, num_repair)

        if not (0 <= op_decay <= 1):
            raise ValueError("op_decay outside [0, 1] not understood.")

        self._op_decay = op_decay

    def update_weights(self, d_idx, r_idx, s_idx):
        self._d_weights[d_idx] *= self._op_decay
        self._d_weights[d_idx] += (1 - self._op_decay) * self._scores[s_idx]

        self._r_weights[r_idx] *= self._op_decay
        self._r_weights[r_idx] += (1 - self._op_decay) * self._scores[s_idx]



from alns.accept.AcceptanceCriterion import AcceptanceCriterion


class HillClimbing(AcceptanceCriterion):
    """
    Hill climbing only accepts progressively better solutions, discarding those
    that result in a worse objective value.
    """

    def __call__(self, rnd, best, current, candidate):
        return candidate.objective() <= current.objective()



from abc import ABC, abstractmethod


class State(ABC):
    """
    State object, which stores a solution via its decision variables. The
    objective value is evaluated via its ``objective()`` member.

    The State class is abstract - you are encouraged to subclass it to suit
    your specific problem.
    """

    @abstractmethod
    def objective(self) -> float:
        """
        Computes the state's associated objective value.

        Returns
        -------
        Some numeric value, e.g. an ``int`` or ``float``.
        """
        return NotImplemented



import numpy as np
import numpy.random as rnd
import pytest
from numpy.testing import assert_

from alns.Result import Result
from alns.Statistics import Statistics
from .states import Sentinel

try:
    from matplotlib.testing.decorators import check_figures_equal
except ImportError:

    def check_figures_equal(*args, **kwargs):  # placeholder
        return check_figures_equal


# HELPERS ---------------------------------------------------------------------


def get_result(state):
    """
    Helper method.
    """
    return Result(state, get_statistics())


def get_statistics():
    """
    Helper method.
    """
    statistics = Statistics()

    for objective in range(100):
        statistics.collect_objective(objective)

    # We should make sure these results are reproducible.
    state = rnd.RandomState(1)

    operators = ["test1", "test2", "test3"]

    for _ in range(100):
        operator = state.choice(operators)

        statistics.collect_destroy_operator("d_" + operator, state.randint(4))
        statistics.collect_repair_operator("r_" + operator, state.randint(4))

    return statistics


# TODO revisit image comparison - maybe check against static images instead?


def get_objective_plot(ax, data, **kwargs):
    """
    Helper method.
    """
    title = kwargs.pop("title", None)

    if title is None:
        title = "Objective value at each iteration"

    ax.plot(data, **kwargs)
    ax.plot(np.minimum.accumulate(data), **kwargs)

    ax.set_title(title)
    ax.set_ylabel("Objective value")
    ax.set_xlabel("Iteration (#)")

    ax.legend(["Current", "Best"], loc="upper right")


def get_operator_plot(
    figure, destroy, repair, legend=None, suptitle=None, **kwargs
):
    """
    Helper method.
    """

    def _helper(ax, operator_counts, title):
        operator_names = list(operator_counts.keys())

        operator_counts = np.array(list(operator_counts.values()))
        cumulative_counts = operator_counts[:, : len(legend)].cumsum(axis=1)

        ax.set_xlim(right=cumulative_counts[:, -1].max())

        for idx in range(len(legend)):
            widths = operator_counts[:, idx]
            starts = cumulative_counts[:, idx] - widths

            ax.barh(operator_names, widths, left=starts, height=0.5, **kwargs)

            for y, (x, label) in enumerate(zip(starts + widths / 2, widths)):
                ax.text(x, y, str(label), ha="center", va="center")

        ax.set_title(title)
        ax.set_xlabel("Iterations where operator resulted in this outcome (#)")
        ax.set_ylabel("Operator")

    if suptitle is not None:
        figure.suptitle(suptitle)

    if legend is None:
        legend = ["Best", "Better", "Accepted", "Rejected"]

    d_ax, r_ax = figure.subplots(nrows=2)

    _helper(d_ax, destroy, "Destroy operators")
    _helper(r_ax, repair, "Repair operators")

    figure.legend(legend, ncol=len(legend), loc="lower center")


# TESTS -----------------------------------------------------------------------


def test_result_state():
    """
    Tests if the result object correctly returns the passed-in state.
    """
    best = Sentinel()

    assert_(get_result(best).best_state is best)


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_objectives(fig_test, fig_ref):
    """
    Tests if the ``plot_objectives`` method returns the same figure as a
    reference plot below.
    """
    result = get_result(Sentinel())

    # Tested plot
    result.plot_objectives(fig_test.subplots())

    # Reference plot
    get_objective_plot(fig_ref.subplots(), result.statistics.objectives)


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_objectives_kwargs(fig_test, fig_ref):
    """
    Tests if the passed-in keyword arguments to ``plot_objectives`` are
    correctly passed to the ``plot`` method.
    """
    result = get_result(Sentinel())
    kwargs = dict(lw=5, marker="*")

    # Tested plot
    result.plot_objectives(fig_test.subplots(), **kwargs)

    # Reference plot
    get_objective_plot(
        fig_ref.subplots(), result.statistics.objectives, **kwargs
    )


@pytest.mark.matplotlib
def test_plot_objectives_default_axes():
    """
    When an axes object is not passed, the ``plot_objectives`` method should
    create a new figure and axes object.
    """
    result = get_result(Sentinel())
    result.plot_objectives()

    # TODO verify the resulting plot somehow


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_operator_counts(fig_test, fig_ref):
    """
    Tests if the ``plot_operator_counts`` method returns the same figure as a
    reference plot below.
    """
    result = get_result(Sentinel())

    # Tested plot
    result.plot_operator_counts(fig_test)

    # Reference plot
    get_operator_plot(
        fig_ref,
        result.statistics.destroy_operator_counts,
        result.statistics.repair_operator_counts,
    )


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_operator_counts_title(fig_test, fig_ref):
    """
    Tests if ``plot_operator_counts`` sets a plot title correctly.
    """
    result = get_result(Sentinel())

    # Tested plot
    result.plot_operator_counts(fig_test, title="A random test title")

    # Reference plot
    get_operator_plot(
        fig_ref,
        result.statistics.destroy_operator_counts,
        result.statistics.repair_operator_counts,
        suptitle="A random test title",
    )


@pytest.mark.matplotlib
def test_plot_operator_counts_default_figure():
    """
    When a figure object is not passed, the ``plot_operator_counts`` method
    should create new figure and axes objects.
    """
    result = get_result(Sentinel())
    result.plot_operator_counts()

    # TODO verify the resulting plot somehow


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_operator_counts_kwargs(fig_test, fig_ref):
    """
    Tests if the passed-in keyword arguments to ``plot_operator_counts`` are
    correctly passed to the ``barh`` method.
    """
    result = get_result(Sentinel())
    kwargs = dict(alpha=0.5, lw=2)

    # Tested plot
    result.plot_operator_counts(fig_test, **kwargs)

    # Reference plot
    get_operator_plot(
        fig_ref,
        result.statistics.destroy_operator_counts,
        result.statistics.repair_operator_counts,
        **kwargs
    )


@pytest.mark.matplotlib
@check_figures_equal(extensions=["png"])
def test_plot_operator_counts_legend_length(fig_test, fig_ref):
    """
    Tests if the length of the passed-in legend is used to determine which
    counts to show.
    """
    result = get_result(Sentinel())

    # Tested plot
    result.plot_operator_counts(fig_test, legend=["Best"])

    # Reference plot
    get_operator_plot(
        fig_ref,
        result.statistics.destroy_operator_counts,
        result.statistics.repair_operator_counts,
        legend=["Best"],
    )



import pytest
from matplotlib import pyplot as plt


@pytest.fixture(autouse=True)
def auto_close_all_figures(request):
    """
    Helper method that closes active testing figures in-between tests. See
    e.g. https://github.com/matplotlib/matplotlib/issues/15079 for details.
    """
    if "matplotlib" in request.keywords:
        plt.close("test")
        plt.close("reference")



import pytest
from numpy.random import RandomState
from numpy.testing import assert_, assert_equal, assert_raises

from alns.stop import MaxIterations
from alns.tests.states import Zero


@pytest.mark.parametrize("max_iterations", [-1, -42, -10000])
def test_raise_negative_parameters(max_iterations: int):
    """
    Maximum iterations cannot be negative.
    """
    with assert_raises(ValueError):
        MaxIterations(max_iterations)


@pytest.mark.parametrize("max_iterations", [1, 42, 10000])
def test_does_not_raise(max_iterations: int):
    """
    Valid parameters should not raise.
    """
    MaxIterations(max_iterations)


@pytest.mark.parametrize("max_iterations", [1, 42, 10000])
def test_max_iterations(max_iterations):
    """
    Test if the max iterations parameter is correctly set.
    """
    stop = MaxIterations(max_iterations)
    assert_equal(stop.max_iterations, max_iterations)


def test_before_max_iterations():
    stop = MaxIterations(100)
    rnd = RandomState(0)

    for _ in range(100):
        assert_(not stop(rnd, Zero(), Zero()))


def test_after_max_iterations():
    stop = MaxIterations(100)
    rnd = RandomState()

    for _ in range(100):
        stop(rnd, Zero(), Zero())

    for _ in range(100):
        assert_(stop(rnd, Zero(), Zero()))



def update(current: float, step: float, method: str) -> float:
    """
    Updates the passed-in criterion threshold parameter. This is done in one of
    two ways, determined via ``method``. If ``method`` is linear, then ``step``
    is subtracted from the threshold. If ``method`` is exponential, the
    threshold is multiplied by ``step``.

    Parameters
    ----------
    current
        The current criterion threshold.
    step
        The chosen step size.
    method
        The updating method, one of {'linear', 'exponential'}.

    Raises
    ------
    ValueError
        When the method is not understood.

    Returns
    -------
    The new criterion threshold.
    """
    method = method.lower()

    if method == "linear":
        return current - step
    elif method == "exponential":
        return current * step

    raise ValueError("Method `{0}' not understood.".format(method))



from typing import Optional

from alns.accept.AcceptanceCriterion import AcceptanceCriterion


class GreatDeluge(AcceptanceCriterion):
    """
    The Great Deluge (GD) criterion accepts solutions if the candidate solution
    has a value lower than a threshold (originally called the water level [1]).
    The initial threshold is computed as

    ``threshold = alpha * initial.objective()``

    where ``initial`` is the initial solution passed-in to ALNS.

    The threshold is updated in each iteration as

    ``threshold = threshold - beta * (threshold - candidate.objective())``

    The implementation is based on the description of GD in [2].

    Parameters
    ----------
    alpha
        Factor used to compute the initial threshold
    beta
        Factor used to update the threshold

    References
    ----------
    [1]: Dueck, G. New optimization heuristics: The great deluge algorithm and
         the record-to-record travel. *Journal of Computational Physics* (1993)
         104 (1): 86-92.
    [2]: Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
         criteria for the adaptive large neighbourhood search metaheuristic.
         *Journal of Heuristics* (2018) 24 (5): 783–815.
    """

    def __init__(self, alpha: float, beta: float):
        if alpha <= 1 or not (0 < beta < 1):
            raise ValueError("alpha must be > 1 and beta must be in (0, 1).")

        self._alpha = alpha
        self._beta = beta
        self._threshold: Optional[float] = None

    @property
    def alpha(self):
        return self._alpha

    @property
    def beta(self):
        return self._beta

    def __call__(self, rnd, best, current, candidate):
        if self._threshold is None:
            self._threshold = self.alpha * best.objective()

        diff = self._threshold - candidate.objective()
        res = diff > 0

        self._threshold -= self.beta * diff

        return res



from abc import ABC, abstractmethod

from numpy.random import RandomState

from alns.State import State


class StoppingCriterion(ABC):
    """
    Base class from which to implement a stopping criterion.
    """

    @abstractmethod
    def __call__(self, rnd: RandomState, best: State, current: State) -> bool:
        """
        Determines whether to stop based on the implemented stopping criterion.

        Parameters
        ----------
        rnd
            May be used to draw random numbers from.
        best
            The best solution state observed so far.
        current
            The current solution state.

        Returns
        -------
        Whether to stop iterating (True), or not (False).
        """
        return NotImplemented



import pytest
from numpy.testing import assert_equal, assert_raises

from alns.accept.update import update


def test_raises_unknown_method():
    with assert_raises(ValueError):
        update(1, 0.5, "unknown_method")

    update(1, 0.5, "linear")  # this should work


@pytest.mark.parametrize(
    "lc,uc",
    [
        ("linear", "LINEAR"),
        ("linear", "Linear"),
        ("exponential", "EXPONENTIAL"),
    ],
)
def test_accepts_any_case_method(lc: str, uc: str):
    """
    ``update`` should be indifferent about the passed-in method case.
    """
    assert_equal(update(1, 0.5, lc), update(1, 0.5, uc))


@pytest.mark.parametrize(
    "curr,step,method,expected",
    [
        (1, 0.5, "linear", 0.5),
        (2, 1, "linear", 1),
        (2, 0, "linear", 2),
        (1, 0.5, "exponential", 0.5),
        (2, 1, "exponential", 2),
        (2, 0, "exponential", 0),
    ],
)
def test_update(curr: float, step: float, method: str, expected: float):
    """
    Tests if linear updating works as designed, as ``current - step``, and
    exponential updating as ``current * step``.
    """
    assert_equal(update(curr, step, method), expected)



import time

import pytest
from numpy.random import RandomState
from numpy.testing import assert_, assert_equal, assert_raises

from alns.stop import MaxRuntime
from alns.tests.states import Zero


def sleep(duration, get_now=time.perf_counter):
    """
    Custom sleep function. Built-in time.sleep function is not precise
    and has different performances depending on the OS, see
    https://stackoverflow.com/questions/1133857/how-accurate-is-pythons-time-sleep
    """
    now = get_now()
    end = now + duration
    while now < end:
        now = get_now()


@pytest.mark.parametrize("max_runtime", [-0.001, -1, -10.1])
def test_raise_negative_parameters(max_runtime: float):
    """
    Maximum runtime may not be negative.
    """
    with assert_raises(ValueError):
        MaxRuntime(max_runtime)


@pytest.mark.parametrize("max_runtime", [0.001, 1, 10.1])
def test_valid_parameters(max_runtime: float):
    """
    Does not raise for non-negative parameters.
    """
    MaxRuntime(max_runtime)


@pytest.mark.parametrize("max_runtime", [0.01, 0.1, 1])
def test_max_runtime(max_runtime):
    """
    Test if the max time parameter is correctly set.
    """
    stop = MaxRuntime(max_runtime)
    assert_equal(stop.max_runtime, max_runtime)


@pytest.mark.parametrize("max_runtime", [0.01, 0.05, 0.10])
def test_before_max_runtime(max_runtime):
    stop = MaxRuntime(max_runtime)
    rnd = RandomState()
    for _ in range(100):
        assert_(not stop(rnd, Zero(), Zero()))


@pytest.mark.parametrize("max_runtime", [0.01, 0.05, 0.10])
def test_after_max_runtime(max_runtime):
    stop = MaxRuntime(max_runtime)
    rnd = RandomState()
    stop(rnd, Zero(), Zero())  # Trigger the first time measurement
    sleep(max_runtime)

    for _ in range(100):
        assert_(stop(rnd, Zero(), Zero()))



from typing import List

import numpy as np
import numpy.random as rnd
from numpy.testing import assert_, assert_raises, assert_almost_equal
from pytest import mark

from alns.weights import SimpleWeights


@mark.parametrize("op_decay", [1.01, -0.01, -0.5, 1.5])
def test_raises_invalid_op_decay(op_decay: float):
    with assert_raises(ValueError):
        SimpleWeights([0, 0, 0, 0], 1, 1, op_decay)


@mark.parametrize("op_decay", np.linspace(0, 1, num=5))
def test_does_not_raise_valid_op_decay(op_decay: float):
    SimpleWeights([0, 0, 0, 0], 1, 1, op_decay)


@mark.parametrize(
    "scores,op_decay,expected",
    [
        ([0, 0, 0, 0], 1, [1, 1]),  # scores are not used
        ([0, 0, 0, 0], 0, [0, 0]),  # initial weights are not used
        ([0.5, 0.5, 0.5, 0.5], 0.5, [0.75, 0.75]),
    ],
)  # convex combination
def test_update_weights(
    scores: List[float], op_decay: float, expected: List[float]
):
    weights = SimpleWeights(scores, 1, 1, op_decay)

    # TODO other weights?
    weights.update_weights(0, 0, 1)

    assert_almost_equal(weights.destroy_weights[0], expected[0])
    assert_almost_equal(weights.repair_weights[0], expected[1])


@mark.parametrize(
    "op_coupling",
    [
        np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]),
        np.array([[1, 0, 1], [0, 1, 1], [1, 1, 0]]),
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]),
        np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]]),
        np.array([[1, 0, 0], [1, 0, 0], [1, 0, 0]]),  # Not allowed by ALNS
    ],
)
def test_select_operators(op_coupling):
    """
    Test if the indices of the selected operators correspond to the
    ones that are given by the operator coupling.
    """
    rnd_state = rnd.RandomState()
    n_destroy, n_repair = op_coupling.shape
    weights = SimpleWeights([0, 0, 0, 0], n_destroy, n_repair, 0)
    d_idx, r_idx = weights.select_operators(rnd_state, op_coupling)

    assert_((d_idx, r_idx) in np.argwhere(op_coupling == 1))



import logging

import numpy as np

from alns.accept.AcceptanceCriterion import AcceptanceCriterion
from alns.accept.update import update

logger = logging.getLogger(__name__)


class SimulatedAnnealing(AcceptanceCriterion):
    """
    Simulated annealing, using an updating temperature. The temperature is
    updated as,

    ``temperature = max(end_temperature, temperature - step)`` (linear)

    ``temperature = max(end_temperature, step * temperature)`` (exponential)

    where the initial temperature is set to ``start_temperature``.

    Parameters
    ----------
    start_temperature
        The initial temperature.
    end_temperature
        The final temperature.
    step
        The updating step.
    method
        The updating method, one of {'linear', 'exponential'}. Default
        'exponential'.

    References
    ----------
    [1]: Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
         criteria for the adaptive large neighbourhood search metaheuristic.
         *Journal of Heuristics* (2018) 24 (5): 783–815.
    [2]: Kirkpatrick, S., Gerlatt, C. D. Jr., and Vecchi, M. P. Optimization
         by Simulated Annealing. *IBM Research Report* RC 9355, 1982.
    """

    def __init__(
        self,
        start_temperature: float,
        end_temperature: float,
        step: float,
        method: str = "exponential",
    ):
        if start_temperature <= 0 or end_temperature <= 0 or step < 0:
            raise ValueError("Temperatures must be strictly positive.")

        if start_temperature < end_temperature:
            raise ValueError(
                "start_temperature < end_temperature not understood."
            )

        if method == "exponential" and step > 1:
            raise ValueError("Exponential updating cannot have step > 1.")

        self._start_temperature = start_temperature
        self._end_temperature = end_temperature
        self._step = step
        self._method = method

        self._temperature = start_temperature

    @property
    def start_temperature(self) -> float:
        return self._start_temperature

    @property
    def end_temperature(self) -> float:
        return self._end_temperature

    @property
    def step(self) -> float:
        return self._step

    @property
    def method(self) -> str:
        return self._method

    def __call__(self, rnd, best, current, candidate):
        probability = np.exp(
            (current.objective() - candidate.objective()) / self._temperature
        )

        # We should not set a temperature that is lower than the end
        # temperature.
        self._temperature = max(
            self.end_temperature,
            update(self._temperature, self.step, self.method),
        )

        # TODO deprecate RandomState in favour of Generator - which uses
        #  random(), rather than random_sample().
        try:
            return probability >= rnd.random()
        except AttributeError:
            return probability >= rnd.random_sample()

    @classmethod
    def autofit(
        cls, init_obj: float, worse: float, accept_prob: float, num_iters: int
    ) -> "SimulatedAnnealing":
        """
        Returns an SA object with initial temperature such that there is a
        ``accept_prob`` chance of selecting a solution up to ``worse`` percent
        worse than the initial solution. The step parameter is then chosen such
        that the temperature reaches 1 in ``num_iters`` iterations.

        This procedure was originally proposed by Ropke and Pisinger (2006),
        and has seen some use since - i.a. Roozbeh et al. (2018).

        Parameters
        ----------
        init_obj
            The initial solution objective.
        worse
            Percentage (between 0 and 1) the candidate solution may be worse
            than initial solution for it to be accepted with probability
            ``accept_prob``.
        accept_prob
            Initial acceptance probability for a solution at most ``worse``
            worse than the initial solution.
        num_iters
            Number of iterations the ALNS algorithm will run.

        Raises
        ------
        ValueError
            When ``worse`` not in [0, 1] or when ``accept_prob``not in (0, 1).

        Returns
        -------
        An autofitted SimulatedAnnealing acceptance criterion.

        References
        ----------
        [1]: Ropke, Stefan, and David Pisinger. 2006. "An Adaptive Large
             Neighborhood Search Heuristic for the Pickup and Delivery Problem
             with Time Windows." _Transportation Science_ 40 (4): 455 - 472.
        [2]: Roozbeh et al. 2018. "An Adaptive Large Neighbourhood Search for
             asset protection during escaped wildfires." _Computers &
             Operations Research_ 97: 125 - 134.
        """
        if not (0 <= worse <= 1):
            raise ValueError("worse outside [0, 1] not understood.")

        if not (0 < accept_prob < 1):
            raise ValueError("accept_prob outside (0, 1) not understood.")

        if num_iters < 0:
            raise ValueError("Negative number of iterations not understood.")

        start_temp = -worse * init_obj / np.log(accept_prob)
        step = (1 / start_temp) ** (1 / num_iters)

        logger.info(f"Autofit start_temp {start_temp:.2f}, step {step:.2f}.")

        return cls(start_temp, 1, step, method="exponential")



from unittest.mock import Mock

import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.accept import WorseAccept
from alns.tests.states import One, Two, Zero


@mark.parametrize(
    "start, end, step, method",
    [
        (-1, 1, 0.1, "linear"),  # start prob cannot be < 0
        (2, 1, 0.1, "linear"),  # start prob cannot be > 1
        (1, -1, 0.1, "linear"),  # end prob cannot be < 0
        (1, 2, 0.1, "linear"),  # end prob cannot be > 1
        (1, 0, -0.1, "linear"),  # updating step cannot be < 0
        (0.5, 0.6, 0.1, "linear"),  # start prob cannot be < end prob
        (1, 0.5, 2, "exponential"),  # step cannot be > 1 with exponential
    ],
)
def test_raises_invalid_parameters(start, end, step, method):
    with assert_raises(ValueError):
        WorseAccept(start, end, step, method)


@mark.parametrize(
    "start, end, step, method",
    [
        (1, 1, 1, "linear"),  # one start and end prob
        (0, 0, 1, "linear"),  # zero start and end prob
        (0.5, 0.5, 1, "linear"),  # equal start and end prob
        (0.05, 0.01, 0.001, "linear"),  # regular start and end prob
        (1, 0.5, 1, "exponential"),  # boundary step exponential
    ],
)
def test_no_raise_valid_parameters(start, end, step, method):
    WorseAccept(start, end, step, method)


@mark.parametrize(
    "start,end,step,method",
    [
        (1, 0, 1, "linear"),
        (0.9, 0.0, 0.1, "linear"),
        (0.5, 0.5, 0.1, "linear"),
        (0, 0, 0, "exponential"),
        (1, 0, 0.9999, "exponential"),
    ],
)
def test_properties(start, end, step, method):
    """
    Tests if the properties are correctly set.
    """
    worse_accept = WorseAccept(start, end, step, method)

    assert_equal(worse_accept.start_prob, start)
    assert_equal(worse_accept.end_prob, end)
    assert_equal(worse_accept.step, step)
    assert_equal(worse_accept.method, method)


def test_zero_prob_accepts_better():
    """
    Tests if WA with a zero start probability accepts better solutions.
    """
    rnd_vals = [1]
    rng = Mock(spec_set=rnd.RandomState, random=lambda: rnd_vals.pop(0))
    worse_accept = WorseAccept(0, 0, 0.1)

    assert_(worse_accept(rng, Zero(), One(), Zero()))
    assert_(worse_accept(rng, Zero(), Two(), Zero()))


def test_zero_prob_never_accept_worse():
    """
    Tests if WA with a zero start probability does not accept worse solutions.
    """
    worse_accept = WorseAccept(0, 0, 0, "linear")

    assert_(not worse_accept(rnd.RandomState(), Zero(), One(), One()))
    assert_(not worse_accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_one_prob_always_accept():
    """
    Tests if WA with a fixed probability of 1 leads to always accepting
    solutions.
    """
    worse_accept = WorseAccept(1, 0, 0, "linear")

    for _ in range(100):
        assert_(worse_accept(rnd.RandomState(), Zero(), Zero(), One()))


def test_linear_consecutive_solutions():
    """
    Test if WA with linear updating method correctly accepts and rejects
    consecutive solutions.
    """
    rnd_vals = [0.9, 0.8, 0.7, 0.6, 0.5, 1]
    rng = Mock(spec_set=rnd.RandomState, random=lambda: rnd_vals.pop(0))
    worse_accept = WorseAccept(1, 0, 0.1, "linear")

    # For the first five, the probability is, resp., 1, 0.9, 0.8, 0.7, 0.6
    # The random draw is, resp., 0.9, 0.8, 0.7, 0.6, 0.5 so the worsening
    # solution is still accepted.
    for _ in range(5):
        assert_(worse_accept(rng, Zero(), Zero(), One()))

    # The probability is now 0.5 and the draw is 1, so reject.
    assert_(not worse_accept(rng, Zero(), Zero(), One()))


def test_exponential_consecutive_solutions():
    """
    Test if WA with exponential updating method correctly accepts and rejects
    consecutive solutions.
    """
    rnd_vals = [0.5, 0.25, 0.125, 1]
    rng = Mock(spec_set=rnd.RandomState, random=lambda: rnd_vals.pop(0))
    worse_accept = WorseAccept(1, 0, 0.5, "exponential")

    # For the first three, the probability is, resp., 1, 0.5, 0.25
    # The random draw is, resp., 0.5, 0.25, 0.125, so the worsening
    # solution is still accepted.
    for _ in range(3):
        assert_(worse_accept(rng, Zero(), Zero(), One()))

    # The probability is now 0.5 and the draw is 1, so reject.
    assert_(not worse_accept(rng, Zero(), Zero(), One()))



import numpy as np
import numpy.random as rnd
from numpy.testing import (
    assert_,
    assert_almost_equal,
    assert_equal,
    assert_raises,
)
from pytest import mark

from alns.accept import SimulatedAnnealing
from alns.tests.states import One, Zero


@mark.parametrize(
    "start,end,step",
    [
        (-1, 1, 1),
        (1, -1, 1),
        (1, 1, -1),
    ],
)
def test_raises_negative_parameters(start: float, end: float, step: float):
    """
    Simulated annealing does not work with negative parameters, so those should
    not be accepted.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(start, end, step)


def test_raises_explosive_step():
    """
    For exponential updating, the step parameter must not be bigger than one,
    as that would result in an explosive threshold.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(2, 1, 2, "exponential")

    SimulatedAnnealing(2, 1, 1, "exponential")  # boundary should be fine


def test_temperature_boundary():
    """
    The boundary case for the end temperature parameter is at zero, which must
    *not* be accepted, as it would result in a division-by-zero.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(1, 0, 1)


def test_raises_start_smaller_than_end():
    """
    The initial temperature at the start should be bigger (or equal) to the end
    temperature.
    """
    with assert_raises(ValueError):
        SimulatedAnnealing(0.5, 1, 1)

    SimulatedAnnealing(1, 1, 1)  # should not raise for equality


def test_does_not_raise():
    """
    These sets of parameters should work correctly.
    """
    SimulatedAnnealing(10, 5, 1, "exponential")
    SimulatedAnnealing(10, 5, 2, "linear")


@mark.parametrize("step", range(10))
def test_step(step: int):
    """
    Tests if the step parameter is correctly set.
    """
    # Linear because exponential does not take explosive step parameters.
    assert_equal(SimulatedAnnealing(1, 1, step, "linear").step, step)


@mark.parametrize("start", range(1, 10))
def test_start_temperature(start: int):
    """
    Tests if the start_temperature parameter is correctly set.
    """
    assert_equal(SimulatedAnnealing(start, 1, 1).start_temperature, start)


@mark.parametrize("end", range(1, 10))
def test_end_temperature(end: float):
    """
    Tests if the end_temperature parameter is correctly set.
    """
    assert_equal(SimulatedAnnealing(10, end, 1).end_temperature, end)


def test_accepts_better():
    for _ in range(1, 100):
        simulated_annealing = SimulatedAnnealing(2, 1, 1)
        assert_(simulated_annealing(rnd.RandomState(), One(), Zero(), Zero()))


def test_accepts_equal():
    simulated_annealing = SimulatedAnnealing(2, 1, 1)

    for _ in range(100):
        # This results in an acceptance probability of exp{0}, that is, one.
        # Thus, the candidate state should always be accepted.
        assert_(simulated_annealing(rnd.RandomState(), One(), One(), One()))


def test_linear_random_solutions():
    """
    Checks if the linear ``accept`` method correctly decides in two known cases
    for a fixed seed.
    """
    simulated_annealing = SimulatedAnnealing(2, 1, 1, "linear")

    state = rnd.RandomState(0)

    # Using the above seed, the first two random numbers are 0.55 and .72,
    # respectively. The acceptance probability is 0.61 first, so the first
    # should be accepted (0.61 > 0.55). Thereafter, it drops to 0.37, so the
    # second should not (0.37 < 0.72).
    assert_(simulated_annealing(state, Zero(), Zero(), One()))
    assert_(not simulated_annealing(state, Zero(), Zero(), One()))


def test_exponential_random_solutions():
    """
    Checks if the exponential ``accept`` method correctly decides in two known
    cases for a fixed seed. This is the exponential equivalent to the linear
    random solutions test above.
    """
    simulated_annealing = SimulatedAnnealing(2, 1, 0.5, "exponential")

    state = rnd.RandomState(0)

    assert_(simulated_annealing(state, Zero(), Zero(), One()))
    assert_(not simulated_annealing(state, Zero(), Zero(), One()))


def test_accepts_generator_and_random_state():
    """
    Tests if SimulatedAnnealing works with both Generator and RandomState
    randomness classes.

    See also https://numpy.org/doc/1.18/reference/random/index.html#quick-start
    """

    class Old:  # old RandomState interface mock
        def random_sample(self):  # pylint: disable=no-self-use
            return 0.5

    simulated_annealing = SimulatedAnnealing(2, 1, 1)
    assert_(simulated_annealing(Old(), One(), One(), Zero()))

    class New:  # new Generator interface mock
        def random(self):  # pylint: disable=no-self-use
            return 0.5

    simulated_annealing = SimulatedAnnealing(2, 1, 1)
    assert_(simulated_annealing(New(), One(), One(), Zero()))


@mark.parametrize(
    "worse,accept_prob,iters",
    [
        (1, 0, 10),  # zero accept prob
        (1, 1.2, 10),  # prob outside unit interval
        (1, 1, 10),  # unit accept prob
        (-1, 0.5, 10),  # negative worse
        (0, -1, 10),  # negative prob
        (1.5, 0.5, 10),  # worse outside unit interval
        (1, 0.9, -10),
    ],
)  # negative number of iterations
def test_autofit_raises_for_invalid_inputs(
    worse: float, accept_prob: float, iters: int
):
    with assert_raises(ValueError):
        SimulatedAnnealing.autofit(1.0, worse, accept_prob, iters)


@mark.parametrize(
    "init_obj,worse,accept_prob,iters",
    [(1_000, 1, 0.9, 1), (1_000, 0.5, 0.05, 1)],
)
def test_autofit_on_several_examples(
    init_obj: float, worse: float, accept_prob: float, iters: int
):
    # We have:
    #   prob = exp{-(f^c - f^i) / T},
    # where T is start temp, f^i the init sol objective, and f^c the candidate
    # solution objective. We also have that f^c is at worst (1 + worse) f^i.
    # Substituting and solving for T, we then find:
    # T = -worse * f^i / ln(p).
    sa_start = -worse * init_obj / np.log(accept_prob)
    sa_end = 1

    # We have end = r ** iters * start, so r = (end / start) ** (1 / iters).
    sa_step = (sa_end / sa_start) ** (1 / iters)

    sa = SimulatedAnnealing.autofit(init_obj, worse, accept_prob, iters)

    assert_almost_equal(sa.start_temperature, sa_start)
    assert_almost_equal(sa.end_temperature, sa_end)
    assert_almost_equal(sa.step, sa_step)
    assert_equal(sa.method, "exponential")



import math

from alns.accept.GreatDeluge import GreatDeluge


class NonLinearGreatDeluge(GreatDeluge):
    """
    The Non-Linear Great Deluge (NLGD) criterion accepts solutions if the
    candidate solution has a value lower than a threshold (originally called
    the water level [1]). The initial threshold is computed as

    ``threshold = alpha * initial.objective()``

    where ``initial`` is the initial solution passed-in to ALNS.

    The non-linear GD variant was proposed by [2]. It differs from GD by using
    a non-linear updating scheme; see the ``_compute_threshold`` method for
    details. Moreover, candidate solutions that improve the current solution
    are always accepted.

    The implementation is based on the description in [2].

    Parameters
    ----------
    alpha
        Factor used to compute the initial threshold. See [2] for details.
    beta
        Factor used to update the threshold. See [2] for details.
    gamma
        Factor used to update the threshold. See [2] for details.
    delta
        Factor used to update the threshold. See [2] for details.

    References
    ----------
    [1]: Dueck, G. New optimization heuristics: The great deluge algorithm and
         the record-to-record travel. *Journal of Computational Physics* (1993)
         104 (1): 86-92.
    [2]: Landa-Silva, D., & Obit, J. H. Great deluge with non-linear decay rate
         for solving course timetabling problems. *4th international IEEE
         conference intelligent systems* (2008) Vol. 1: 8-11.
    [3]: Santini, A., Ropke, S. & Hvattum, L.M. A comparison of acceptance
         criteria for the adaptive large neighbourhood search metaheuristic.
         *Journal of Heuristics* (2018) 24 (5): 783–815.
    """

    def __init__(self, alpha: float, beta: float, gamma: float, delta: float):
        super().__init__(alpha, beta)

        if gamma <= 0 or delta <= 0:
            raise ValueError("Gamma and delta must be non-negative.")

        self._gamma = gamma
        self._delta = delta

    @property
    def gamma(self):
        return self._gamma

    @property
    def delta(self):
        return self._delta

    def __call__(self, rnd, best, current, candidate):
        if self._threshold is None:
            if best.objective() == 0:
                raise ValueError("Initial solution cannot have zero value.")

            self._threshold = self._alpha * best.objective()

        res = candidate.objective() < self._threshold

        if not res:
            # Accept if improving
            res = candidate.objective() < current.objective()

        self._threshold = self._compute_threshold(best, current, candidate)

        return res

    def _compute_threshold(self, best, curr, cand):
        """
        Returns the new threshold value.

        First, the relative gap between the candidate solution and threshold
        is computed. If this relative gap is less than ``beta``, then the
        threshold is linearly increased (involving the ``gamma`` parameter).
        Otherwise, the threshold is exponentially decreased (involving the
        ``delta`` parameter).
        """
        rel_gap = (self._threshold - cand.objective()) / self._threshold

        if rel_gap < self._beta:
            term1 = self.gamma * abs(cand.objective() - self._threshold)
            term2 = self._threshold
        else:
            term1 = self._threshold * math.exp(-self.delta * best.objective())
            term2 = best.objective()

        return term1 + term2



import time

from typing import Optional
from numpy.random import RandomState

from alns.State import State
from alns.stop.StoppingCriterion import StoppingCriterion


class MaxRuntime(StoppingCriterion):
    """
    Criterion that stops after a specified maximum runtime.
    """

    def __init__(self, max_runtime: float):
        if max_runtime < 0:
            raise ValueError("max_runtime < 0 not understood.")

        self._max_runtime = max_runtime
        self._start_runtime: Optional[float] = None

    @property
    def max_runtime(self) -> float:
        return self._max_runtime

    def __call__(self, rnd: RandomState, best: State, current: State) -> bool:
        if self._start_runtime is None:
            self._start_runtime = time.perf_counter()

        return time.perf_counter() - self._start_runtime > self.max_runtime



from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
from numpy.random import RandomState


class WeightScheme(ABC):
    """
    Base class from which to implement a weight scheme.

    Parameters
    ----------
    scores
        A list of four non-negative elements, representing the weight
        updates when the candidate solution results in a new global best
        (idx 0), is better than the current solution (idx 1), the solution
        is accepted (idx 2), or rejected (idx 3).
    num_destroy
        Number of destroy operators.
    num_repair
        Number of repair operators.
    """

    def __init__(self, scores: List[float], num_destroy: int, num_repair: int):
        self._validate_arguments(scores, num_destroy, num_repair)

        self._scores = scores
        self._d_weights = np.ones(num_destroy, dtype=float)
        self._r_weights = np.ones(num_repair, dtype=float)

    @property
    def destroy_weights(self) -> np.ndarray:
        return self._d_weights

    @property
    def repair_weights(self) -> np.ndarray:
        return self._r_weights

    def select_operators(
        self, rnd_state: RandomState, op_coupling: np.ndarray
    ) -> Tuple[int, int]:
        """
        Selects a destroy and repair operator pair to apply in this iteration.
        The default implementation uses a roulette wheel mechanism, where each
        operator is selected based on the normalised weights.

        Parameters
        ----------
        rnd_state
            Random state object, to be used for random number generation.
        op_coupling
            Matrix that indicates coupling between destroy and repair
            operators. Entry (i, j) is 1 if destroy operator i can be used in
            conjunction with repair operator j and 0 otherwise.

        Returns
        -------
        A tuple of (d_idx, r_idx), which are indices into the destroy and
        repair operator lists, respectively.
        """

        def select(op_weights):
            probs = op_weights / np.sum(op_weights)
            return rnd_state.choice(range(len(op_weights)), p=probs)

        d_idx = select(self._d_weights)
        coupled_r_idcs = np.flatnonzero(op_coupling[d_idx])
        r_idx = coupled_r_idcs[select(self._r_weights[coupled_r_idcs])]

        return d_idx, r_idx

    @abstractmethod
    def update_weights(self, d_idx: int, r_idx: int, s_idx: int):
        """
        Updates the weights associated with the applied destroy (d_idx) and
        repair (r_idx) operators. The score index (s_idx) indicates the
        outcome.

        Parameters
        ----------
        d_idx
            Destroy operator index.
        r_idx
            Repair operator index
        s_idx
            Score index.
        """
        return NotImplemented

    @staticmethod
    def _validate_arguments(scores, num_destroy, num_repair):
        if any(score < 0 for score in scores):
            raise ValueError("Negative scores are not understood.")

        if len(scores) < 4:
            # More than four is OK because we only use the first four.
            raise ValueError(f"Expected four scores, found {len(scores)}")

        if num_destroy <= 0 or num_repair <= 0:
            raise ValueError("Missing destroy or repair operators.")



from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.pyplot import Axes, Figure

from alns.State import State
from alns.Statistics import Statistics


class Result:
    """
    Stores ALNS results. An instance of this class is returned once the
    algorithm completes.

    Parameters
    ----------
    best
        The best state observed during the entire iteration.
    statistics
        Statistics collected during iteration.
    """

    def __init__(self, best: State, statistics: Statistics):
        self._best = best
        self._statistics = statistics

    @property
    def best_state(self) -> State:
        """
        The best state observed during the entire iteration.
        """
        return self._best

    @property
    def statistics(self) -> Statistics:
        """
        The statistics object populated during iteration.
        """
        return self._statistics

    def plot_objectives(
        self,
        ax: Optional[Axes] = None,
        title: Optional[str] = None,
        **kwargs: Dict[str, Any]
    ):
        """
        Plots the collected objective values at each iteration.

        Parameters
        ----------
        ax
            Optional axes argument. If not passed, a new figure and axes are
            constructed.
        title
            Optional title argument. When not passed, a default is set.
        kwargs
            Optional arguments passed to ``ax.plot``.
        """
        if ax is None:
            _, ax = plt.subplots()

        if title is None:
            title = "Objective value at each iteration"

        # First call is current solution objectives (at each iteration), second
        # call is the best solution found so far (as a running minimum).
        ax.plot(self.statistics.objectives, **kwargs)
        ax.plot(np.minimum.accumulate(self.statistics.objectives), **kwargs)

        ax.set_title(title)
        ax.set_ylabel("Objective value")
        ax.set_xlabel("Iteration (#)")

        ax.legend(["Current", "Best"], loc="upper right")

        plt.draw_if_interactive()

    def plot_operator_counts(
        self,
        fig: Optional[Figure] = None,
        title: Optional[str] = None,
        legend: Optional[List[str]] = None,
        **kwargs: Dict[str, Any]
    ):
        """
        Plots an overview of the destroy and repair operators' performance.

        Parameters
        ----------
        fig
            Optional figure. If not passed, a new figure is constructed, and
            some default margins are set.
        title
            Optional figure title. When not passed, no title is set.
        legend
            Optional legend entries. When passed, this should be a list of at
            most four strings. The first string describes the number of times
            a best solution was found, the second a better, the third a
            solution was accepted but did not improve upon the current or
            global best, and the fourth the number of times a solution was
            rejected. If less than four strings are passed, only the first
            len(legend) count types are plotted. When not passed, a sensible
            default is set and all counts are shown.
        kwargs
            Optional arguments passed to each call of ``ax.barh``.
        """
        if fig is None:
            fig, (d_ax, r_ax) = plt.subplots(nrows=2)
            fig.subplots_adjust(hspace=0.7, bottom=0.2)
        else:
            d_ax, r_ax = fig.subplots(nrows=2)

        if title is not None:
            fig.suptitle(title)

        if legend is None:
            legend = ["Best", "Better", "Accepted", "Rejected"]

        self._plot_op_counts(
            d_ax,
            self.statistics.destroy_operator_counts,
            "Destroy operators",
            min(len(legend), 4),
            **kwargs
        )

        self._plot_op_counts(
            r_ax,
            self.statistics.repair_operator_counts,
            "Repair operators",
            min(len(legend), 4),
            **kwargs
        )

        fig.legend(legend[:4], ncol=len(legend), loc="lower center")

        plt.draw_if_interactive()

    @staticmethod
    def _plot_op_counts(ax, operator_counts, title, num_types, **kwargs):
        """
        Internal helper that plots the passed-in operator_counts on the given
        ax object.

        Note
        ----
        This code takes loosely after an example from the matplotlib gallery
        titled "Discrete distribution as horizontal bar chart".
        """
        operator_names = list(operator_counts.keys())

        operator_counts = np.array(list(operator_counts.values()))
        cumulative_counts = operator_counts[:, :num_types].cumsum(axis=1)

        ax.set_xlim(right=cumulative_counts[:, -1].max())

        for idx in range(num_types):
            widths = operator_counts[:, idx]
            starts = cumulative_counts[:, idx] - widths

            ax.barh(operator_names, widths, left=starts, height=0.5, **kwargs)

            for y, (x, label) in enumerate(zip(starts + widths / 2, widths)):
                ax.text(x, y, str(label), ha="center", va="center")

        ax.set_title(title)
        ax.set_xlabel("Iterations where operator resulted in this outcome (#)")
        ax.set_ylabel("Operator")



import numpy as np
import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.accept import GreatDeluge
from alns.tests.states import One, Two, Zero


@mark.parametrize(
    "alpha, beta",
    [
        (1, 0.5),  # alpha cannot be <= 1
        (2, 1),  # beta cannot be outside of (0, 1)
    ],
)
def test_raise_invalid_parameters(alpha, beta):
    with assert_raises(ValueError):
        GreatDeluge(alpha, beta)


@mark.parametrize("alpha, beta", [(1.01, 0.01), (999, 0.5)])
def test_no_raise_valid_parameters(alpha, beta):
    GreatDeluge(alpha, beta)


@mark.parametrize("alpha", np.arange(1.1, 10, 1))
def test_alpha(alpha):
    great_deluge = GreatDeluge(alpha, 0.1)
    assert_equal(great_deluge.alpha, alpha)


@mark.parametrize("beta", np.arange(0.1, 1, 0.1))
def test_beta(beta):
    great_deluge = GreatDeluge(2, beta)
    assert_equal(great_deluge.beta, beta)


def test_accepts_below_threshold():
    great_deluge = GreatDeluge(2.01, 0.5)

    # Initial threshold is set at 2.01, hence Two should be accepted
    assert_(great_deluge(rnd.RandomState(), One(), Zero(), Two()))


def test_rejects_above_threshold():
    great_deluge = GreatDeluge(1.01, 0.5)

    # Initial threshold is set at 1.01, hence Two should be rejected
    assert_(not great_deluge(rnd.RandomState(), One(), Zero(), Two()))


def test_rejects_equal_threshold():
    great_deluge = GreatDeluge(2, 0.5)

    # Initial threshold is set at 2, hence Two should be rejected
    assert_(not great_deluge(rnd.RandomState(), One(), Zero(), Two()))


def test_evaluate_consecutive_solutions():
    """
    Test if GD correctly accepts and rejects consecutive solutions.
    """
    great_deluge = GreatDeluge(2, 0.01)

    # The initial threshold is set at 2*0 = 0, so the first candidate with
    # value one is rejected. The threshold is updated to 0.01.
    assert_(not great_deluge(rnd.RandomState(), Zero(), Zero(), One()))

    # The second candidate is below the threshold (0 < 0.01), hence accepted.
    # The threshold is updated to 0.0099.
    assert_(great_deluge(rnd.RandomState(), Zero(), Zero(), Zero()))

    # The third candidate is below the threshold (0 < 0.0099), hence accepted.
    assert_(great_deluge(rnd.RandomState(), Zero(), Zero(), Zero()))



from alns import State


class VarObj(State):
    """Test solution state object with variable objective."""

    def __init__(self, obj: float):
        self.obj = obj

    def objective(self) -> float:
        return self.obj


Sentinel = lambda: VarObj(0)  # noqa: E731
Zero = lambda: VarObj(0)  # noqa: E731
One = lambda: VarObj(1)  # noqa: E731
Two = lambda: VarObj(2)  # noqa: E731



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alns',
 'alns.accept',
 'alns.accept.tests',
 'alns.stop',
 'alns.stop.tests',
 'alns.tests',
 'alns.weights',
 'alns.weights.tests']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=2.2.0', 'numpy>=1.15.2']

setup_kwargs = {
    'name': 'alns',
    'version': '4.1.0',
    'description': 'A flexible implementation of the adaptive large neighbourhood search (ALNS) algorithm.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/alns.svg)](https://badge.fury.io/py/alns)\n[![ALNS](https://github.com/N-Wouda/ALNS/actions/workflows/alns.yml/badge.svg)](https://github.com/N-Wouda/ALNS/actions/workflows/alns.yml)\n[![codecov](https://codecov.io/gh/N-Wouda/ALNS/branch/master/graph/badge.svg)](https://codecov.io/gh/N-Wouda/ALNS)\n\nThis package offers a general, well-documented and tested\nimplementation of the adaptive large neighbourhood search (ALNS)\nmeta-heuristic, based on the description given in [Pisinger and Ropke\n(2010)][1]. It may be installed in the usual way as\n```\npip install alns\n```\n\n### Examples\nIf you wish to dive right in, the `examples/` directory contains example notebooks\nshowing how the ALNS library may be used. These include:\n\n- The travelling salesman problem (TSP), [here][2]. We solve an\n  instance of 131 cities to within 2% of optimality, using simple\n  destroy and repair heuristics with a post-processing step.\n- The cutting-stock problem (CSP), [here][4]. We solve an instance with\n  180 beams over 165 distinct sizes to within 1.35% of optimality in\n  only a very limited number of iterations.\n- The resource-constrained project scheduling problem, [here][6]. We solve an\n  instance with 90 jobs and 4 resources to within 4% of the best known solution,\n  using a number of different operators and enhancement techniques from the \n  literature.\n\nFinally, the features notebook gives an overview of various options available \nin the `alns` package (explained below). In the notebook we use these different\noptions to solve a toy 0/1-knapsack problem. The notebook is a good starting\npoint for when you want to use different schemes, acceptance or stopping criteria \nyourself. It is available [here][5].\n\n## How to use\nThe `alns` package exposes two classes, `ALNS` and `State`. The first\nmay be used to run the ALNS algorithm, the second may be subclassed to\nstore a solution state - all it requires is to define an `objective`\nmember function, returning an objective value.\n\nThe ALNS algorithm must be supplied with a _weight scheme_, an _acceptance\ncriterion_, and a _stopping criterion_.\n\n### Weight scheme\nThe weight scheme determines how to select destroy and repair operators in each\niteration of the ALNS algorithm. Several have already been implemented for you,\nin `alns.weights`:\n\n- `SimpleWeights`. This weight scheme applies a convex combination of the \n   existing weight vector, and a reward given for the current candidate \n   solution.\n- `SegmentedWeights`. This weight scheme divides the iteration horizon into\n   segments. In each segment, scores are summed for each operator. At the end\n   of each segment, the weight vector is updated as a convex combination of \n   the existing weight vector, and these summed scores.\n\nEach weight scheme inherits from `WeightScheme`, which may be used to write \nyour own.\n\n### Acceptance criterion\nThe acceptance criterion determines the acceptance of a new solution state at\neach iteration. An overview of common acceptance criteria is given in\n[Santini et al. (2018)][3]. Several have already been implemented for you, in\n`alns.accept`. These include:\n\n- `HillClimbing`. The simplest acceptance criterion, hill-climbing solely\n  accepts solutions improving the objective value.\n- `RecordToRecordTravel`. This criterion accepts solutions when the improvement\n  meets some updating threshold.\n- `SimulatedAnnealing`. This criterion accepts solutions when the\n  scaled probability is bigger than some random number, using an\n  updating temperature.\n- And many more!\n\nEach acceptance criterion inherits from `AcceptanceCriterion`, which may be used\nto write your own.\n\n### Stoppping criterion\nThe stopping criterion determines when ALNS should stop iterating. Several \ncommonly used stopping criteria have already been implemented for you, in\n`alns.stop`:\n\n- `MaxIterations`. This stopping criterion stops the heuristic search after a\n  given number of iterations.\n- `MaxRuntime`. This stopping criterion stops the heuristic search after a given\n  number of seconds.\n\nEach stopping criterion inherits from `StoppingCriterion`, which may be used to\nwrite your own.\n\n## References\n- Pisinger, D., and Ropke, S. (2010). Large Neighborhood Search. In M.\n  Gendreau (Ed.), _Handbook of Metaheuristics_ (2 ed., pp. 399-420).\n  Springer.\n- Santini, A., Ropke, S. & Hvattum, L.M. (2018). A comparison of\n  acceptance criteria for the adaptive large neighbourhood search\n  metaheuristic. *Journal of Heuristics* 24 (5): 783-815.\n\n[1]: http://orbit.dtu.dk/en/publications/large-neighborhood-search(61a1b7ca-4bf7-4355-96ba-03fcdf021f8f).html\n[2]: https://github.com/N-Wouda/ALNS/blob/master/examples/travelling_salesman_problem.ipynb\n[3]: https://link.springer.com/article/10.1007%2Fs10732-018-9377-x\n[4]: https://github.com/N-Wouda/ALNS/blob/master/examples/cutting_stock_problem.ipynb\n[5]: https://github.com/N-Wouda/ALNS/blob/master/examples/alns_features.ipynb\n[6]: https://github.com/N-Wouda/ALNS/blob/master/examples/resource_constrained_project_scheduling_problem.ipynb\n',
    'author': 'Niels Wouda',
    'author_email': 'n.wouda@apium.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/N-Wouda/ALNS',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod

from numpy.random import RandomState

from alns.State import State


class AcceptanceCriterion(ABC):
    """
    Base class from which to implement an acceptance criterion.
    """

    @abstractmethod
    def __call__(
        self, rnd: RandomState, best: State, current: State, candidate: State
    ) -> bool:
        """
        Determines whether to accept the proposed, candidate solution based on
        this acceptance criterion and the other solution states.

        Parameters
        ----------
        rnd
            May be used to draw random numbers from.
        best
            The best solution state observed so far.
        current
            The current solution state.
        candidate
            The proposed solution state.

        Returns
        -------
        Whether to accept the candidate state (True), or not (False).
        """
        return NotImplemented



import numpy.random as rnd
from numpy.testing import (
    assert_,
    assert_allclose,
    assert_equal,
    assert_raises,
)
from pytest import mark

from alns.accept import RecordToRecordTravel
from alns.tests.states import One, Zero


@mark.parametrize(
    "start, end, step, method",
    [
        (-1, 1, 1, "linear"),  # start threshold cannot be negative
        (1, -1, 1, "linear"),  # nor can the end threshold
        (1, 1, -1, "linear"),  # nor the updating step
        (0, 1, 1, "linear"),  # start threshold cannot be < end threshold
        (2, 1, 2, "exponential"),  # step cannot be > 1 with exponential
        (1, 1, 1, "non-linear"),  # method must be 'linear' or 'exponential'
    ],
)
def test_raises_invalid_parameters(start, end, step, method):
    with assert_raises(ValueError):
        RecordToRecordTravel(start, end, step, method)


@mark.parametrize(
    "start, end, step, method",
    [
        (0, 0, 1, "linear"),  # boundary start and end threshold
        (0.5, 0.5, 1, "linear"),  # non-boundary equal start and end
        (10, 0, 0, "linear"),  # boundary step
        (1, 0.5, 1, "exponential"),  # boundary step exponential
        (1000, 0.5, 0.1, "linear"),  # regular parameters
    ],
)
def test_no_raise_valid_parameters(start, end, step, method):
    RecordToRecordTravel(start, end, step, method)


@mark.parametrize(
    "start, end, step, method",
    [
        (1, 0, 1, "linear"),
        (0.5, 0.5, 1, "linear"),
        (10, 5, 0.001, "linear"),
        (1, 0.5, 1, "exponential"),
    ],
)
def test_properties(start, end, step, method):
    """
    Test if the properties are correctly set.
    """
    rrt = RecordToRecordTravel(start, end, step, method)
    assert_equal(rrt.start_threshold, start)
    assert_equal(rrt.end_threshold, end)
    assert_equal(rrt.step, step)
    assert_equal(rrt.method, method)


def test_accepts_better():
    record_travel = RecordToRecordTravel(1, 0, 0.1)
    assert_(record_travel(rnd.RandomState(), One(), Zero(), Zero()))


def test_rejects_worse():
    record_travel = RecordToRecordTravel(0.5, 0.2, 0.1)

    # This results in a relative worsening of plus one, which is bigger than
    # the threshold (0.5).
    assert_(not record_travel(rnd.RandomState(), Zero(), Zero(), One()))


def test_accepts_equal():
    record_travel = RecordToRecordTravel(0, 0, 0.1)

    # Even a the strictest threshold, this should be accepted since the
    # relative improvement is zero (they are equal).
    assert_(record_travel(rnd.RandomState(), Zero(), Zero(), Zero()))


def test_linear_threshold_update():
    record_travel = RecordToRecordTravel(5, 0, 1)

    # For the first five, the threshold is, resp., 5, 4, 3, 2, 1. The relative
    # worsening is plus one, so this should be accepted (lower or equal to
    # threshold).
    for _ in range(5):
        assert_(record_travel(rnd.RandomState(), Zero(), Zero(), One()))

    # Threshold is now zero, so this should no longer be accepted.
    assert_(not record_travel(rnd.RandomState(), Zero(), Zero(), One()))


def test_exponential_threshold_update():
    record_travel = RecordToRecordTravel(5, 0, 0.1, "exponential")

    # The relative worsening is plus one, so this should be accepted initially,
    # as the threshold is 5, resp. 0.5. In the second case, 1 > 0.5, so the
    # second should be rejected.
    assert_(record_travel(rnd.RandomState(), Zero(), Zero(), One()))
    assert_(not record_travel(rnd.RandomState(), Zero(), Zero(), One()))


@mark.parametrize(
    "init_obj, start_gap, end_gap, n_iters, method",
    [
        (1, -1, 0, 1, "linear"),  # negative start
        (1, 0, -1, 1, "linear"),  # negative end
        (1, 0.5, 1, 1, "linear"),  # start < end
        (1, 0, 0, 0, "linear"),  # non-positive n_iters
        (1, 0, 0, 1, "nonlinear"),  # invalid method
    ],
)
def test_autofit_raises_for_invalid_inputs(
    init_obj: float,
    start_gap: float,
    end_gap: float,
    n_iters: int,
    method: str,
):
    with assert_raises(ValueError):
        RecordToRecordTravel.autofit(
            init_obj, start_gap, end_gap, n_iters, method
        )


@mark.parametrize(
    "init_obj,start_gap,end_gap,n_iters,method,exp_start,exp_end,exp_step",
    [
        (1, 1, 0, 1, "linear", 1, 0, 1),
        (10, 0.1, 0, 10, "linear", 1, 0, 0.1),
        (100, 0.5, 0.5, 10, "linear", 50, 50, 0),
        (1, 1, 0.1, 1, "exponential", 1, 0.1, 0.1),
        (10, 0.1, 0.01, 10, "exponential", 1, 0.1, 0.79432),
        (100, 0.05, 0.05, 10, "exponential", 5, 5, 1),
    ],
)
def test_autofit_on_several_examples(
    init_obj: float,
    start_gap: float,
    end_gap: float,
    n_iters: int,
    method: str,
    exp_start: float,
    exp_end: float,
    exp_step: float,
):
    rrt = RecordToRecordTravel.autofit(
        init_obj, start_gap, end_gap, n_iters, method
    )

    assert_allclose(rrt.start_threshold, exp_start)
    assert_allclose(rrt.end_threshold, exp_end)
    assert_allclose(rrt.step, exp_step, rtol=1e-3)
    assert_equal(rrt.method, method)



import logging
from typing import List

import numpy as np

from alns.weights.WeightScheme import WeightScheme

logger = logging.getLogger(__name__)


class SegmentedWeights(WeightScheme):
    """
    A segmented weight scheme. Weights are not updated in each iteration,
    but only after each segment. Scores are gathered during each segment,
    as:

    ``seg_weight += score``

    At the start of each segment, ``seg_weight`` is reset to zero. At the
    end of a segment, the weights are updated as:

    ``new_weight = seg_decay * old_weight + (1 - seg_decay) * seg_weight``

    Parameters
    ----------
    (other arguments are explained in ``WeightScheme``)

    seg_decay
        Decay parameter in [0, 1]. This parameter is used to weigh segment
        and overall performance of each operator.
    seg_length
        Length of a single segment. Default 100.
    """

    def __init__(
        self,
        scores: List[float],
        num_destroy: int,
        num_repair: int,
        seg_decay: float,
        seg_length: int = 100,
    ):
        super().__init__(scores, num_destroy, num_repair)

        if not (0 <= seg_decay <= 1):
            raise ValueError("seg_decay outside [0, 1] not understood.")

        if seg_length < 1:
            raise ValueError("seg_length < 1 not understood.")

        self._seg_decay = seg_decay
        self._seg_length = seg_length
        self._iter = 0

        self._reset_segment_weights()

    def select_operators(self, rnd_state, op_coupling):
        self._iter += 1

        if self._iter % self._seg_length == 0:
            logger.debug(f"End of segment (#iters = {self._iter}).")

            self._d_weights *= self._seg_decay
            self._d_weights += (1 - self._seg_decay) * self._d_seg_weights

            self._r_weights *= self._seg_decay
            self._r_weights += (1 - self._seg_decay) * self._r_seg_weights

            self._reset_segment_weights()

        return super().select_operators(rnd_state, op_coupling)

    def update_weights(self, d_idx, r_idx, s_idx):
        self._d_seg_weights[d_idx] += self._scores[s_idx]
        self._r_seg_weights[r_idx] += self._scores[s_idx]

    def _reset_segment_weights(self):
        self._d_seg_weights = np.zeros_like(self._d_weights)
        self._r_seg_weights = np.zeros_like(self._r_weights)



import numpy.random as rnd
from numpy.testing import assert_, assert_equal, assert_raises
from pytest import mark

from alns.accept import LateAcceptanceHillClimbing
from alns.tests.states import One, Two, Zero


@mark.parametrize("lookback_period", [-0.01, -10, 1.5])
def test_raises_invalid_lookback_period(lookback_period):
    with assert_raises((ValueError, TypeError)):
        LateAcceptanceHillClimbing(lookback_period=lookback_period)


@mark.parametrize(
    "lookback_period, greedy, better_history",
    [
        (0, True, True),
        (1, False, True),
        (10, True, False),
        (100, False, False),
    ],
)
def test_properties(lookback_period, greedy, better_history):
    lahc = LateAcceptanceHillClimbing(lookback_period, greedy, better_history)

    assert_equal(lahc.lookback_period, lookback_period)
    assert_equal(lahc.better_history, better_history)
    assert_equal(lahc.greedy, greedy)


@mark.parametrize(
    "greedy, better_history",
    [(False, False), (True, False), (False, True), (True, True)],
)
def test_zero_lookback_period(greedy, better_history):
    """
    Test if LAHC behaves like regular hill climbing when `lookback_period` is
    set to zero.
    """
    lahc = LateAcceptanceHillClimbing(0, greedy, better_history)

    assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))
    assert_(not lahc(rnd.RandomState(), Zero(), One(), One()))
    assert_(not lahc(rnd.RandomState(), Zero(), Zero(), Zero()))
    assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))


@mark.parametrize("lookback_period", [0, 3, 10, 50])
def test_accept(lookback_period):
    """
    Tests if LAHC accepts a solution that is better than the current solution
    from `lookback_period` iterations ago.
    """
    lahc = LateAcceptanceHillClimbing(lookback_period, False, False)

    assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))

    for _ in range(lookback_period):
        # The then-current solution `lookback_period` iterations ago is 2, so
        # the candidate solution with value 1 should be accepted.
        assert_(lahc(rnd.RandomState(), Zero(), One(), One()))


@mark.parametrize("lookback_period", [0, 3, 10, 50])
def test_reject(lookback_period):
    """
    Tests if LAHC rejects a solution that is worse than the current solution
    from `lookback_period` iterations ago.
    """
    lahc = LateAcceptanceHillClimbing(lookback_period, False, False)

    for _ in range(lookback_period):
        assert_(lahc(rnd.RandomState(), Zero(), One(), Zero()))

    # The then-current solution from `lookback_period` iterations ago has
    # value 1, so the candidate solution with value 1 should be rejected.
    assert_(not lahc(rnd.RandomState(), Zero(), Zero(), One()))


@mark.parametrize("lookback_period", [0, 3, 10, 50])
def test_greedy_accept(lookback_period):
    """
    Tests that if `greedy` is set, then a solution that is better than
    the current solution is accepted, despite being worse than the current
    solution from `lookback_period` iterations ago.
    """
    lahc = LateAcceptanceHillClimbing(lookback_period, True, False)

    assert_(not lahc(rnd.RandomState(), Zero(), Zero(), Two()))

    for _ in range(lookback_period):
        # The candidate solution (1) is accepted because it is better than the
        # current solution (2), despite being worse than the compared
        # historial solution from `lookback_period` iterations ago (1).
        assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))


def test_better_history_small_example():
    """
    Tests if only current solutions are stored that are better than
    the compared previous solution when `better_history` is set.
    """
    lahc = LateAcceptanceHillClimbing(1, False, True)

    assert_(lahc(rnd.RandomState(), Zero(), One(), Zero()))
    assert_(lahc(rnd.RandomState(), Zero(), Two(), Zero()))

    # Previous current stays at 1 because 2 was not better
    assert_(not lahc(rnd.RandomState(), Zero(), Zero(), One()))

    # Previous current is updated to Zero
    assert_(not lahc(rnd.RandomState(), Zero(), Zero(), Zero()))


@mark.parametrize("lookback_period", [3, 10, 50])
def test_better_history_reject(lookback_period):
    """
    Tests that if `better_history` is set, a solution can be rejected despite
    being better than the actual current solution from `lookback_period`
    iterations ago.
    """
    lahc = LateAcceptanceHillClimbing(lookback_period, False, True)

    for _ in range(lookback_period):
        assert_(not lahc(rnd.RandomState(), Zero(), One(), Two()))

    for _ in range(lookback_period):
        # The current solutions are not stored because they are worse
        # than the historial current solutions
        assert_(not lahc(rnd.RandomState(), Zero(), Two(), Two()))

    for _ in range(lookback_period):
        # The candidates (1) do not improve the historical solutions (1)
        assert_(not lahc(rnd.RandomState(), Zero(), Two(), One()))


def test_full_example():
    """
    A full example to illustrate the LAHC criterion with a single
    lookback_period, greedy acceptance and better history management.
    """
    lahc = LateAcceptanceHillClimbing(1, True, True)

    # The first iteration compares candidate (1) against current (2)
    # and stores current in the history.
    assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))

    # The second candidate (1) is accepted based on late-acceptance (2).
    # The historical value is updated to 1.
    assert_(lahc(rnd.RandomState(), Zero(), One(), One()))

    # The third candidate (1) is accepted based on greedy comparison with
    # the current solution (2). The historical value is not updated because
    # the current does not improve the historical value (1).
    assert_(lahc(rnd.RandomState(), Zero(), Two(), One()))

    # The fourth candidate (1) is not accepted because it does not improve
    # the current (1) nor does it improve the historical value (1).
    assert_(not lahc(rnd.RandomState(), Zero(), One(), One()))
